from __future__ import print_function

from kabaret import flow

from .action_context import ActionContextGroup

class NoDialogAction(flow.Action):

    def needs_dialog(self):
        return False

    def run(self, button):
        print('Well done Jolly Jumper !')

class SimpleDialogAction(flow.Action):

    def needs_dialog(self):
        return True

    def get_buttons(self):
        self.message.set('This is the action message: <3')
        return ['Ok', 'Maybe', 'Nope']

    def run(self, button):
        print('You chose {}'.format(button))

class DialogParamsAction(flow.Action):

    cut_down_trees = flow.BoolParam(True)
    skip_and_jump = flow.BoolParam(True)
    press_wild_flowers = flow.BoolParam(True)

    def needs_dialog(self):
        return True

    def get_buttons(self):
        self.message.set('Do you:')
        return ['Ok']

    def run(self, button):
        all_true = (
            self.cut_down_trees.get()
            and self.skip_and_jump.get()
            and self.press_wild_flowers.get()
        )
        if (
            self.cut_down_trees.get()
            and self.skip_and_jump.get()
            and self.press_wild_flowers.get()
        ):
            print('You hang around in bars')
        else:
            print('Do you ever party ?')

class DispatchParams(flow.Object):

    pool = flow.Param('Farm')
    priority = flow.Param(50)

    def get_flags(self):
        return [
            '-P', str(self.pool.get()),
            '-p', str(self.priority.get()),
        ]


class SequenceParam(flow.Object):

    first_frame = flow.Param(1)
    last_frame = flow.Param(100)

    def get_flags(self):
        return [
            '--first', str(self.first_frame.get()),
            '--last', str(self.last_frame.get()),
        ]

class ComplexDialogAction(flow.Action):

    dispatch = flow.Child(DispatchParams)
    sequence = flow.Child(SequenceParam)

    def needs_dialog(self):
        return True

    def get_buttons(self):
        self.message.set('Configure and Submit your job')
        return ['Submit']

    def run(self, button):
        cmd = ['spam_it']+self.dispatch.get_flags()+self.sequence.get_flags()
        print('#---> Cmd:', ' '.join(cmd))

class KeepOpenDialogAction(flow.Action):

    def needs_dialog(self):
        return True

    def get_buttons(self):
        self.message.set('Click the Ok button !')
        return ['Ok', 'Close']

    def run(self, button):
        if button == 'Close':
            return

        self.message.set('Alrigth, now click the Close button :)')
        return self.get_result(close=False)

class GotoAction(flow.Action):

    _parent = flow.Parent()

    # def __init__(self, *args, **kwargs):
    #     super(GotoAction, self).__init__(*args, **kwargs)
    #     self._target_name = None

    def needs_dialog(self):
        return True

    def get_buttons(self):
        self._target_name = None
        return ['Goto Project', 'Goto Parent', 'Create Page and use it for all Goto', 'Open new a page every time',]

    def run(self, button):
        if button == 'Goto Project':
            project_oid = self.root().project().oid()
            return self.get_result(goto=project_oid, close=False, goto_target=self._target_name)

        if button == 'Goto Parent':
            parent_oid = self._parent.oid()
            return self.get_result(goto=parent_oid, close=False, goto_target=self._target_name)

        if button == 'Create Page and use it for all Goto':
            project_oid = self.root().project().oid()
            import time
            self._target_name = str(time.time())
            return self.get_result(goto=project_oid, close=False, goto_target=self._target_name)

        if button == 'Open new a page every time':
            project_oid = self.root().project().oid()
            return self.get_result(goto=project_oid, close=False, goto_target='_NEW_')


class EditMyValueAction(flow.Action):

    _value = flow.Parent()

    def needs_dialog(self):
        return True

    def get_buttons(self):
        self.message.set('Pick a color:')
        return ['Red', 'Pink', 'Blue', 'Violet']

    def run(self, button):
        self._value.set(button)

class SetMyValueAction(flow.Action):

    _value = flow.Parent()

    def needs_dialog(self):
        return False

    def run(self, button):
        self._value.set('orange')

class MyValue(flow.values.Value):

    edit = flow.Child(EditMyValueAction)
    orange = flow.Child(SetMyValueAction)

class SubDialogAction(flow.Action):

    quote = flow.Param('', MyValue)

    def needs_dialog(self):
        return True

    def run(self, button):
        print('You selected: "{}"'.format(self.quote.get()))

class WizardDialogPage1(flow.Action):

    _parent = flow.Parent()

    def needs_dialog(self):
        return True

    def get_buttons(self):
        self.message.set('This is page <B>1/3</B> of a wizard style action')
        return ['Next']

    def run(self, button):
        return self.get_result(next_action=self._parent.wizard_page_2.oid())

class WizardDialogPage2(flow.Action):

    _parent = flow.Parent()

    def needs_dialog(self):
        return True

    def get_buttons(self):
        self.message.set('This is page <B>2/3</B> of a wizard style action')
        return ['Prev', 'Next']

    def run(self, button):
        if button == 'Next':
            return self.get_result(next_action=self._parent.wizard_page_3.oid())
        else:
            return self.get_result(next_action=self._parent.wizard_page_1.oid())

class WizardDialogPage3(flow.Action):

    _parent = flow.Parent()

    def needs_dialog(self):
        return True

    def get_buttons(self):
        self.message.set('This is page <B>3/3</B> of a wizard style action')
        return ['Prev', 'Finish']

    def run(self, button):
        if button == 'Prev':
            return self.get_result(next_action=self._parent.wizard_page_2.oid())
        print('#---> You are a Wizard now !!!')


class ActionsGroup(flow.Object):

    no_dialog_action = flow.Child(NoDialogAction)
    simple_dialog_action = flow.Child(SimpleDialogAction)
    keep_dialog_open_action = flow.Child(KeepOpenDialogAction)
    goto_action = flow.Child(GotoAction)
    dialog_params_action = flow.Child(DialogParamsAction)
    complex_dialog_action = flow.Child(ComplexDialogAction)

    sub_dialog_action = flow.Child(SubDialogAction)

    wizard_page_1 = flow.Child(WizardDialogPage1)
    wizard_page_2 = flow.Child(WizardDialogPage2).ui(hidden=True)
    wizard_page_3 = flow.Child(WizardDialogPage3).ui(hidden=True)

    sep = flow.Label(
        '<hr>'
        '<h2>Action Context</h2>'
        'Action are accessible from various places in the GUI. '
        'Each Action class can specify if it wants to appear in a given context by '
        'implementing the <b>allow_context(context)</b> method. Returning True will let '
        'the GUI show the action, returning False will prevent the GUI from considering '
        'the action. Note that even if shown, the action may be disabled by calling '
        'ui(enabled=False) on the relation pointing to it with<br> '
        'The context argument value will depend of the GUI context gathering the actions: '
        '<ul>'
        '<li>'
            'In Flow view\'s "inline" object representation:<br> '
            'A down arrow button is show at the right of Objects. '
            'It shows every Child action of the Object. '
            'This context is <font color=orange><B>Flow.inline</B></font><br>'
            'Note that this menu may in some case show actions from grand-children. '
            'In such cases, the context has a number of ">" appended: '
            '<font color=orange><B>Flow.inline&gt;&gt;&gt;</B></font><br>'
            'The number of ">" corresponds to the depth from the Object showing '
            'the inline representation.'
        '</li>'
        '<li>'
            'In Flow view\'s "details" object representation:<br> '
            'A button is show inside Objects for each of its '
            'Children actions. '
            'This context is <font color=orange><B>Flow.details</B></font> '
        '</li>'
        '<li>'
            'In Flow view\'s mapped items representation:<br> '
            'A right-click shows an action menu. '
            'This context is <font color=orange><B>Flow.map</B></font> '
        '</li>'
        '<li>'
            'Other view\'s may display action. Most views will use a context '
            'named after the view_type. For example, the ScriptView uses the '
            '<font color=orange><b>Script</b></font> context.'
        '</li>'
        '<li>'
            'The context may be <font color=orange><b>None</b></font> with legacy view '
            'code. In this case you can choose to return None so that the legacy '
            'behavior is used (using the class <i>SHOW_IN_PARENT_INLINE</i> and '
            '<i>SHOW_IN_PARENT_DETAILS</i> attributes) '
            'You should not relay on this since it will be deprecated at some point. '
        '</li>'
        '</ul>'
        'Enter the group below to explore a set of examples:'
    )
    action_context = flow.Child(ActionContextGroup)




import os

from qtpy import QtGui, QtWidgets

from .. import Style


class DarkStyle(Style):

    def __init__(self, name=None):
        super(DarkStyle, self).__init__(name or 'Dark')

    def apply(self, widget):
        app = QtWidgets.QApplication.instance()

        # --- Stuff we could not deal with only using css:

        self.set_property('alternate_child_color', False)

        # --- Change palette only for app wide apply:
    
        if widget is app:
            # setup the palette
            palette = QtWidgets.QApplication.palette()

            # A general background color.
            palette.setColor(QtGui.QPalette.Window, QtGui.QColor(60, 60, 60))

            # A general foreground color.
            palette.setColor(QtGui.QPalette.WindowText,
                             QtGui.QColor(185, 185, 185))

            # Used mostly as the background color for text entry widgets, but can also be used for other painting -
            # such as the background of combobox drop down lists and toolbar handles. It is usually white or another light color.
            palette.setColor(QtGui.QPalette.Base, QtGui.QColor(42, 42, 42))

            # Used as the alternate background color in views with alternating row colors (see QAbstractItemView.setAlternatingRowColors()).
            palette.setColor(QtGui.QPalette.AlternateBase,
                             QtGui.QColor(64, 64, 64))

            # Used as the background color for QToolTip and QWhatsThis. Tool tips use the Inactive color group of QPalette, because tool tips are not active windows.
            palette.setColor(QtGui.QPalette.ToolTipBase, QtGui.QColor(117, 77, 77))

            # Used as the foreground color for QToolTip and QWhatsThis. Tool tips use the Inactive color group of QPalette, because tool tips are not active windows.
            palette.setColor(QtGui.QPalette.ToolTipText, QtGui.QColor(255, 0, 0))

            # The foreground color used with Base. This is usually the same as the WindowText, in which case it must provide good contrast with Window and Base.
            palette.setColor(QtGui.QPalette.Text, QtGui.QColor(128, 128, 128))

            # The general button background color. This background can be different from Window as some styles require a different background color for buttons.
            palette.setColor(QtGui.QPalette.Button, QtGui.QColor(64, 64, 64))

            # A foreground color used with the Button color.
            palette.setColor(QtGui.QPalette.ButtonText,
                             QtGui.QColor(185, 185, 185))

            # A text color that is very different from WindowText, and contrasts well with e.g. Dark. Typically used for text that needs to be drawn where Text
            # or WindowText would give poor contrast, such as on pressed push buttons. Note that text colors can be used for things other than just words; text
            # colors are usually used for text, but it's quite common to use the text color roles for lines, icons, etc.
            palette.setColor(QtGui.QPalette.BrightText,
                             QtGui.QColor(255, 255, 255))

            # Lighter than Button color.
            palette.setColor(QtGui.QPalette.Light, QtGui.QColor(200, 200, 200))

            # Between Button and Light.
            palette.setColor(QtGui.QPalette.Midlight, QtGui.QColor(100, 100, 100))

            # Between Button and Dark.
            palette.setColor(QtGui.QPalette.Mid, QtGui.QColor(40, 40, 40))

            # Darker than Button.
            palette.setColor(QtGui.QPalette.Dark, QtGui.QColor(25, 25, 25))

            # A very dark color. By default, the shadow color is Qt.black.
            palette.setColor(QtGui.QPalette.Shadow, QtGui.QColor(10, 10, 0))

            # A color to indicate a selected item or the current item. By default, the highlight color is Qt.darkBlue.
            palette.setColor(QtGui.QPalette.Highlight, QtGui.QColor(80, 0, 0))

            # A text color that contrasts with Highlight. By default, the highlighted text color is Qt.white.
            palette.setColor(QtGui.QPalette.HighlightedText,
                             QtGui.QColor(188, 128, 128))

            QtWidgets.QApplication.setPalette(palette)

            # Qt Style
            QtWidgets.QApplication.setStyle(QtWidgets.QStyleFactory.create("Plastique"))

        # --- Load and apply the css

        this_folder = os.path.dirname(__file__)
        css_file = os.path.join(this_folder, 'dark_style.css')
        with open(css_file, 'r') as r:
            css = r.read()
        widget.setStyleSheet(css)





import logging
import os
import glob


_PATHS = {}
_CACHE = {}
_PIXCACHE = {}
_ICONCACHE = {}


class ResourcesError(Exception):
    pass


class NotFoundError(ResourcesError):
    pass


def add_folder(name, path):
    '''
    Register a folder to the search path.

    The name parameter will be used to retrieve files.

    The path parameter can be a file inside the folder
    to register, or the path of the folder itself.

    If a folder with the same name was previously added, the
    files from the last one overrides the previous one but
    the previous file not available in the last one are
    still accessible. (i.e. the last one overrides and
    extends the previous one.)

    Every file in this folder will be available with:
        import kabaret.app.resources
        kabaret.app.resources.get(folder_name, file_name)
    (See get(), get_pixmap(), ...)

    Typical usage:
        - add resources from a python package:
            import kabaret.app.resources
            kabaret.app.resources.add_folder('icons', __file__)
        - use resources from this package:
            import kabaret.app.resources
            kabaret.app.resources.get('icons', 'maya')

    '''
    global _PATHS, _CACHE

    if not os.path.isdir(path):
        path = os.path.dirname(path)

    if name not in _PATHS:
        _PATHS[name] = []
    _PATHS[name].insert(0, path)

    # Remove cache for this folder name:
    if name in _CACHE:
        _CACHE.pop(name)


def get_folder_names():
    '''
    Returns a list of available folder names.
    '''
    global _PATHS
    return list(_PATHS)


def list_folder(folder_name):
    '''
    Returns a list of resource names available in the given folder_name.
    '''
    global _PATHS

    EXCLUDE_LIST = ('Thumbs.db',)

    folder_name = folder_name.lower()

    ret = set()
    for path in _PATHS[folder_name]:
        ret.update(
            [
                name.split('.', 1)[0]
                for name in os.listdir(path)
                if not name.startswith('_')
                if not name.startswith('.')
                and name not in EXCLUDE_LIST
            ]
        )
    return sorted(ret)


def list_folder_paths(folder_name):
    '''
    Return a list of list of file_name for the given folder_name.
    (overrides come first)
    '''
    global _PATHS

    folder_name = folder_name.lower()

    ret = []
    for path in _PATHS[folder_name]:
        ret.append(sorted(
            [
                name.split('.', 1)[0]
                for name in os.listdir(path)
                if not name.startswith('_')
            ]
        ))
    return ret


def get(folder_name, file_name):
    '''
    Returns the file named file_name and registered
    under the folder folder_name.
    '''
    global _PATHS, _CACHE

    # avoid case sensitive look-up in order to have the same
    # behavior in win$ and Linux:
    folder_name = folder_name.lower()
    file_name = file_name.lower()

    cache_key = (folder_name, file_name)
    if cache_key in _CACHE:
        if _CACHE[cache_key]:
            return _CACHE[cache_key]
        else:
            raise NotFoundError("Resource not found: " + repr(cache_key))

    try:
        path_list = _PATHS[folder_name]
    except KeyError:
        raise ResourcesError(
            "Unknown resource folder " +
            repr(folder_name) + ". (Did you install it first?)"
        )

    files = None
    for path in path_list:
        files = glob.glob(
            os.path.join(
                path,
                file_name + ('.' not in file_name and ".*" or "")
            )
        )
        if files:
            break
    if not files:
        _CACHE[cache_key] = None
        raise NotFoundError("Resource not found: " + repr(cache_key))
    if len(files) > 1:
        logging.getLogger('kabaret.resources').warn(
            "Several resources found for: " + repr(cache_key))

    _CACHE[cache_key] = files[0]
    return files[0]


def get_pixmap(folder_name, pixmap_name):
    '''
    Same as get, but returns a QPixmap
    '''
    global _PIXCACHE

    path = get(folder_name, pixmap_name)
    if path in _PIXCACHE:
        return _PIXCACHE[path]

    from qtpy import QtGui

    ret = QtGui.QPixmap(path)
    _PIXCACHE[path] = ret

    return ret


def get_icon(icon_ref, for_widget=None, disabled_ref=None):
    '''
    If icon ref is an int, for_widget must not be None
    and its current QStyle will be used to return
    the Qt icon pointed by icon_ref.

    If icon_ref is a 2D tuple, it is used to call
    get_pixmap(*icon_ref)

    If icon_ref is a file path, an icon is created from this file.

    if disable_ref is not None, it must be a 2D tuple suitable for
    get_pixmap(*disabled_ref) and will be used as the "disabled"
    state of the icon.

    Warning: using different disabled_ref for the same icon_ref has 
    undefined behavior.

    '''
    if isinstance(icon_ref, int):
        style = for_widget.style()
        return style.standardIcon(icon_ref)

    global _ICONCACHE
    if icon_ref in _ICONCACHE:
        return _ICONCACHE[icon_ref]

    from qtpy import QtGui

    pix = None
    try:
        pix = get_pixmap(*icon_ref)
    except Exception:
        try:
            if os.path.isfile(icon_ref):
                pix = QtGui.QPixmap(icon_ref)
        except TypeError:
            pass

    if pix is None:
        raise NotFoundError('Cannot find icon for %r' % (icon_ref,))

    icon = QtGui.QIcon(pix)

    if disabled_ref is not None:
        pix = get_pixmap(*disabled_ref)
        icon.addPixmap(pix, mode=icon.Disabled)

    _ICONCACHE[icon_ref] = icon
    return icon



'''

    Lets do this:
        FLOW
        - really granular queries for flow gui
        - hierarchical flow gui view (open a related page inside the page, with indentation)
        - find a way to have some relations on the same line (use relation ui_config infos)
        - file listener editor (for log files)
        - chat editor (for message maps)
        - Map gui should update mapped item infos when it receives a touch event for on of its mapped
        items (not refresh the whole map)
        - flow views should use current object.get_source_display() as view title ! :)
        - ICON in flow object should be able to point to alternative styles/resources repo
        
        PROCESS ACTOR
        - responsible for spawning typed subprocesses (editors)
        - defines env configs for each editor
        - logs outputs for all runs, has a view to show them
        - let the user kill subprocesses ?

        - One must be able to override/augment editor infos (exec_path, env vars,...) before spawining one.
            (the flow will need that)

        GUI:
        - get ride of the default scrolled content
        
'''


import os
import sys
import logging

from kabaret.app.ui.gui.embedded import MayaEmbeddedSession

from kabaret.app.ui.gui.widgets.flow.flow_view import FlowView, FlowContextView


class MyStudioMayaSession(MayaEmbeddedSession):

    def register_view_types(self):
        type_name = self.main_window_manager.register_view_type(FlowView)
        #self.main_window_manager.add_view(type_name)

        type_name = self.main_window_manager.register_view_type(
            FlowContextView
        )
        self.main_window_manager.add_view(type_name)


def install_kabaret():
    logging.getLogger('kabaret.demo').info('ACTUAL INSTALL')

    session = MyStudioMayaSession(session_name="MyStudioMaya")
    session.cmds.Cluster.connect_from_env()
    session.start()
    sys.kabaret_session = session   # What... Yes I can :p



import kabaret.app.resources as resources
resources.add_folder('icons.flow', __file__)

# Some Icons are:
# Icons made by Dave Gandy from www.flaticon.com, licensed by CC 3.0 BY (http://creativecommons.org/licenses/by/3.0/)
# <div>Icons made by <a href="https://www.flaticon.com/authors/dave-gandy" title="Dave Gandy">Dave Gandy</a> from <a href="https://www.flaticon.com/" title="Flaticon">www.flaticon.com</a> is licensed by <a href="http://creativecommons.org/licenses/by/3.0/" title="Creative Commons BY 3.0" target="_blank">CC 3.0 BY</a></div>






import time
import math

from kabaret.flow import (
    values,
    Object, Map, Action, ConnectAction,
    ChoiceValueSetAction, ChoiceValueSelectAction,
    Child, Parent, Computed, Connection,
    Label, Param, SessionParam,
    IntParam, BoolParam, FloatParam, StringParam,
    DictParam, OrderedStringSetParam, HashParam,
)


class StaticChoices(values.ChoiceValue):

    CHOICES = ['Static', 'List', 'Of', 'Choices', 'NYS',
               'INV', 'WAIT', 'WIP', 'RVW', 'RTK', 'DONE', 'OOP']


class DynamicChoices(values.ChoiceValue):

    def choices(self):
        choices = {
            None: ('This is a', 'Dynamic list', 'Of choices'),
            'This is a': ('I can changes', 'depending on', 'what you want', '[reset]'),
            'Dynamic list': ('So you can', 'implement', 'workflow statuses', '[reset]'),
            'Of Choices': ('Voila...', '[reset]'),
            '[reset]': ('This is a', 'Dynamic list', 'Of choices'),
        }
        value = self.get()
        return choices.get(value, ['[reset]'])


class ComputedExample(Object):

    warning = Label(
        '<hr><h2>'
        'By default, Computed values are updated on every get() '
        'and the value is not stored in the value store.<br>'
        '<font color=#f44>/!\\ You cannot assume the computation will append the minimum number of times !</font><br>'
        'See how refreshing the page recompute "Value" (and see how the compute_count increments too).'
    )

    value = Computed().watched()
    compute_count = IntParam(0)

    def compute_child_value(self, child_value):
        # FIXME: this is call twice for each GUI page load :/
        # find why and fix it, or uncached Computed relation make no sens :(
        if child_value == self.value:
            self.value.set(time.time())
            self.compute_count.incr()


class TouchComputedExample(Object):

    help = Label(
        '<hr><h2>'
        'Change the "First" or "Last" param to see the "Count" being recomputed.<br>'
        'See how refreshing the page does not affect the "Computed Count".'
    )

    first = Param(1).watched()
    last = Param(100).watched()
    count = Computed(cached=True)
    computed_count = IntParam(0)

    def child_value_changed(self, child):
        if child in (self.first, self.last):
            self.count.touch()

    def compute_child_value(self, child_value):
        if child_value is self.count:
            self.count.set(
                int(self.last.get()) - int(self.first.get()) + 1
            )
            self.computed_count.incr()


class ResetMyIntAction(Action):
    _value = Parent()

    def get_buttons(self):
        return ['Confirm']

    def run(self, button):
        self._value.set(0)


class MyIntWithAction(values.IntValue):

    reset = Child(ResetMyIntAction)


class ValuesGroup(Object):

    param_info = Label(
        '<hr><h2>'
        'Values are what configures your Flow, '
        'either by User input or by Computation.<br>'
        'There are several subclasses. Some ensure type checking, other add '
        'functionnality like the ChoiceValue.<br>'
        'You add a Value to an Object using the "Param" relation.'
        '(or any of its subclasses)'
        '</h2>'
    )

    default_info = Label(
        '<hr><H3>'
        'The default is to accept any value.<br>Here you can enter anything:'
    )
    default_value = Param([1, 2, 3])

    session_info = Label(
        '<HR><H3>'
        'A "Session" value exists only in memory (not saved to the value store) and '
        'will return to its default for every session.'
    )
    session_value = SessionParam('Default Value...')

    typed_info = Label(
        '<HR><h3>'
        'There are several typed Value, each one with a corresponding Param subclasse and '
        'default editor:'
    )
    int_value = IntParam(0)
    bool_value = BoolParam(True)
    float_value = FloatParam(math.pi)
    string_value = StringParam('Yippie Ki Yay !')
    dict_value = DictParam({1: 11, 'a': 'aa'})
    ordered_string_set_value = OrderedStringSetParam()
    hash_value = HashParam()

    custom_info = Label(
        '<HR><h3>'
        'Param subclasses are mostly just preconfigured Param. You can achieve the same result by '
        'specifying the value type:'
    )
    my_int = Param(909, values.IntValue)
    custom_example = Label(
        '<HR><h3>'
        'This is the way to go if you need to use your very own value type<br>'
        'Here is a Value with actions attached to it:'
    )
    my_int_with_action = Param(909, MyIntWithAction)

    choice_info = Label(
        '<HR><h3>'
        'The ChoiceValue does not have a corresponding Param subclasse since you will need to '
        'inherit it to specify the choices.<br>'
        'Here are two examples:'
    )
    static_choices = Param(None, StaticChoices)
    dynamic_choices = Param(None, DynamicChoices)

    compute_info = Label(
        '<HR><h3>'
        'Some value can be computed instead of input from user, here are examples.</h3>'
        '(Open those pages to test refresh behavior)'
    )
    computed = Child(ComputedExample)
    touch_computed = Child(TouchComputedExample)





from kabaret import flow

USE_INGRID = True
try:
    from kabaret.ingrid.flow_objects import InGridConfig
except ImportError:
    USE_INGRID = False

class DOAction(flow.Action):

    def allow_context(self, context):
        return context and context.endswith('.details')

    def needs_dialog(self):
        return False

class DetailOnlyActions(flow.Object):

    label = flow.Label(
        'All actions here only show on this Object details. '
        '(there is no action menu next to the Object field name)'
    )
    action_1 = flow.Child(DOAction)
    action_2 = flow.Child(DOAction)
    action_3 = flow.Child(DOAction)


class IOAction(flow.Action):

    def allow_context(self, context):
        return context and context.endswith('.inline')

    def needs_dialog(self):
        return False

class InlineOnlyActions(flow.Object):

    label = flow.Label(
        'All actions here only show on this Object inline. '
        '(there is no action menu inside the Object)'
    )

    action_1 = flow.Child(IOAction)
    action_2 = flow.Child(IOAction)
    action_3 = flow.Child(IOAction)


class DAIAction(flow.Action):

    def allow_context(self, context):
        return True

    def needs_dialog(self):
        return False

class DetailAndInlineActions(flow.Object):

    label = flow.Label(
        'All actions here are visible inline and in details. '
    )

    action_1 = flow.Child(DAIAction)
    action_2 = flow.Child(DAIAction)
    action_3 = flow.Child(DAIAction)


class InGridAction(flow.Action):

    def allow_context(self, context):
        return context and context.startswith('InGrid')

    def needs_dialog(self):
        return False

class InGridRowAction(InGridAction):

    def allow_context(self, context):
        return context == 'InGrid.row'

class InGridColumnAction(InGridAction):

    def allow_context(self, context):
        return context == 'InGrid.column'

class InGridCellAction(InGridAction):

    def allow_context(self, context):
        return context == 'InGrid.cell'

class TestInGridConfig(InGridConfig):

    _parent = flow.Parent()

    def get_oid_with_actions(self):
        return self._parent.oid()

    def get_rows(self):
        oid = self.get_oid_with_actions()
        return [
            dict(display=str(i), oid=oid)
            for i in range(20)
        ]

    def get_columns(self):
        oid = self.get_oid_with_actions()
        return [
            dict(display=c, oid=oid)
            for c in 'ABCDEFGHIJKLMNOPQ'
        ]

    def get_cell(self, row, column):
        oid = self.get_oid_with_actions()
        return dict(
            display='{}x{}'.format(row, column),
            oid=oid,
            bg=None,
            fg=None,
            icon=None,
            ui=None,
            align='center', # one of ['left', 'right', 'center']
        )

    # def get_actions(self, *row_column_pairs):
    #     '''
    #     Return a list of (label, action_oid) for the given list
    #     of (row, cols)
    #     '''
    #     return []

class InGridOnlyActions(flow.Object):

    label = flow.Label(
        'Actions here are only visible inside a InGid view. '
        'You will need to have <i>kabaret.ingrid</i> extension installed to test this.<br>'
        '(Drop this group on a InGrid view to load it)<br>'
        'Some are only visible in grid headers, some in grid cells.'
    )
    row_only_action = flow.Child(InGridRowAction)
    column_only_action = flow.Child(InGridColumnAction)
    cell_action = flow.Child(InGridCellAction)
    ingrid_action = flow.Child(InGridAction)

    sep = flow.Separator()  # there's a layout bug, this is a workaround

    _ingrid_config = flow.Child(TestInGridConfig)

    def get_ingrid_config_oids(self):
        return [self._ingrid_config.oid()]

class ActionContextGroup(flow.Object):

    detail_only_actions = flow.Child(DetailOnlyActions)
    inline_only_actions = flow.Child(InlineOnlyActions)
    detail_and_inline_actions = flow.Child(DetailAndInlineActions)
    ingrid_only_actions = flow.Child(InGridOnlyActions)





import os
import six
from qtpy import QtWidgets, QtCore

from ...icons import flow  # noqa # to be sure flow icons are registered.

#!!from ..view import View
from ..widget_view import DockedView, DialogView

from .. import event_filters

from .flow_form import FlowForm
from .navigator import Navigator
from .navigation_control import (
    NavigationOIDControls,
    NavigationHistoryControls,
    NavigationBar
)
from .script_line import ScriptLine


class FlowPage(QtWidgets.QWidget):

    def __init__(self, parent, view, start_oid, root_oid):
        super(FlowPage, self).__init__(parent)

        self.view = view
        self.session = view.session

        self._navigator = Navigator(
            self.session, root_oid, start_oid
        )
        self._navigator.set_create_view_function(view.create_view)

        self.nav_bar = NavigationBar(self, self._navigator)
        self.nav_ctrl = self.nav_bar.nav_ctrl
        self.nav_oid = self.nav_bar.nav_oid

        self.form = FlowForm(self, self)

        lo = QtWidgets.QVBoxLayout()
        lo.addWidget(self.nav_bar)
        lo.setContentsMargins(0, 0, 0, 0)
        lo.setSpacing(0)
        self.setLayout(lo)
        lo.addWidget(self.form, 100)

        self._navigator.add_on_current_changed(self.refresh)

        self._source_view_id = None
        # self._dialogs = {}  # oid to dialg widget

        #FIXME: wtf is this ?!? why isn't this in NavigationBar class ?!?
        self.nav_bar.setAttribute(QtCore.Qt.WA_StyledBackground, True)

    def root_oid(self):
        return self._navigator.root_oid()

    def current_oid(self):
        return self._navigator.current_oid()

    def set_source_view_id(self, source_view_id):
        '''
        Used by run_action to find affected vie by stuffs like 'goto' request
        '''
        self._source_view_id = source_view_id

    def open(self, oid):
        self._navigator.goto(oid, new_view=True)

    def goto(self, oid):
        in_new_view = (
            QtWidgets.QApplication.keyboardModifiers() == QtCore.Qt.ControlModifier
        )
        self._navigator.goto(oid, in_new_view)

    def goto_connected(self, oid):
        in_new_view = (
            QtWidgets.QApplication.keyboardModifiers() == QtCore.Qt.ControlModifier
        )
        self._navigator.goto_parent_of(oid, in_new_view)

    def show_action_dialog(self, action_oid):
        if not self.session.cmds.Flow.action_needs_dialog(action_oid):
            self.run_action(action_oid, None)
            return

        # try:
        #     dialog = self._dialogs[action_oid]
        # except:
        #     pass
        # else:
        #     dialog.show()
        #     dialog.raise_()
        #     dialog.activateWindow()
        #     return

        dialog = FlowDialogView(
            self.session, 
            view_id=None, 
            source_view_id=self.view.view_id(), 
            parent_widget=self.view, 
            start_oid=action_oid, 
            root_oid=action_oid,
            form_config=self.form.config()
        )

        # def on_dialog_close(result, page=self, oid=action_oid):
        #     # dialog = page._dialogs.pop(oid)
        #     dialog.forget_dialog_view()
        #     dialog.deleteLater()
        # dialog.finished.connect(on_dialog_close)

        # self._dialogs[action_oid] = dialog

        dialog.show()

    def run_action(self, action_oid, button):
        # if from_dialog and self._master_page is not None:
        #     self._master_page.run_action(action_oid, button)
        #     return

        # if from_dialog:
        #     dialog = self._dialogs[action_oid]
        #     page = dialog.flow_page
        # else:
        #     dialog = None
        #     page = self

        result = self.session.cmds.Flow.run_action(action_oid, button)
        result = result or {}

        if result.get('refresh', False):
            self.refresh()

        goto_oid = result.get('goto')
        if goto_oid is not None:
            target = result.get('goto_target', self._source_view_id)
            if target is None:
                target = self.view.view_id()
            view_type_name = result.get('goto_target_type', 'Flow')
            view = self.session.find_view(
                view_type_name=view_type_name, 
                view_id=target
            )
            if view is None:
                new_view_id = None
                if target != '_NEW_':
                    new_view_id = target
                view = self.session.add_view(
                    view_type_name=view_type_name, 
                    view_id=new_view_id
                )
            view.goto_request(goto_oid)

        next_oid = result.get('next_action')
        if next_oid is not None:
            self.goto(next_oid)
            return

        if result.get('close', True):
            self.view.on_action_result_close()

    def clear(self):
        self.form.clear()

    def refresh(self):
        oid = self.current_oid()
        view_title = self.session.cmds.Flow.get_source_display(oid)
        self.view.set_view_title(view_title)

        self.clear()

        self.nav_oid.update()
        self.nav_ctrl.update()

        self.form.build_roots(oid)

    def receive_event(self, event, data):
        # print('   --- FlowPage got event', event, data)
        if event == 'flow_touched':
            oid = data['oid']
            root = self.form.invisibleRootItem()
            for i in range(root.childCount()):
                item = root.child(i)
                item.on_touch_event(oid=oid)
            # for oid, dialog in six.iteritems(self._dialogs):
            #     page = dialog.flow_page
            #     page.receive_event(event, data)


class FlowDialogView(DialogView):

    def __init__(self, session, view_id, parent_widget, source_view_id, start_oid, root_oid, form_config):
        super(FlowDialogView, self).__init__(session, view_id, parent_widget)

        self.setWindowTitle(root_oid)

        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

        self.flow_page = FlowPage(self, self, start_oid, root_oid)
        self.flow_page.set_source_view_id(source_view_id)
        self.flow_page.form.configure(**form_config)

        self.layout().addWidget(self.flow_page)

        sg = QtWidgets.QSizeGrip(self)
        sg.setMinimumSize(5,5)
        self.layout().addWidget(sg, 0, QtCore.Qt.AlignRight)

        self.installEventFilter(event_filters.EatEnterKeys())
        self.flow_page.form.viewport().installEventFilter(
            event_filters.MouseResizer(self)
        )        

    def goto_request(self, oid):
        self.flow_page.goto(oid)

    def create_view(self, oid=None):
        # Needed as being the view of a FlowPage()
        pass

    def fitSize(self):
        self.flow_page.form.fitSize()
        self.flow_page.adjustSize()
        self.adjustSize()

    def show(self):
        super(FlowDialogView, self).show()
        self.flow_page.refresh()
        self.fitSize()

    def receive_event(self, event, data):
        # print('   --- FlowDialogView got event', event, data)
        if event == 'flow_touched':
            self.flow_page.receive_event(event, data)

    def on_action_result_close(self):
        self.accept()

# class XX_FlowDialog(QtWidgets.QDialog):

#     def __init__(self, master_page, parent, view, start_oid, root_oid):
#         super(FlowDialog, self).__init__(parent)

#         self.setWindowTitle(root_oid)

#         self.setLayout(QtWidgets.QVBoxLayout())
#         self.layout().setContentsMargins(0, 0, 0, 0)

#         self.flow_page = FlowPage(self, view, start_oid, root_oid)
#         self.flow_page.set_master_page(master_page)
#         self.flow_page.form.configure(**master_page.form.config())

#         self.layout().addWidget(self.flow_page)

#         sg = QtWidgets.QSizeGrip(self)
#         sg.setMinimumSize(5,5)
#         self.layout().addWidget(sg, 0, QtCore.Qt.AlignRight)

#         self.installEventFilter(event_filters.EatEnterKeys())
#         self.flow_page.form.viewport().installEventFilter(
#             event_filters.MouseResizer(self)
#         )

#         self.setProperty("dialog", True)

#     def fitSize(self):
#         self.flow_page.form.fitSize()
#         self.flow_page.adjustSize()
#         self.adjustSize()

#     def show(self):
#         super(FlowDialog, self).show()
#         self.flow_page.refresh()
#         self.fitSize()

class FlowView(DockedView):

    @classmethod
    def view_type_name(cls):
        return 'Flow'

    def __init__(self, session, view_id=None, hidden=False, area=None, oid=None, root_oid=None):
        self._start_oid = oid
        self._root_oid = root_oid
        self.dev_menu = None
        self.script_line = None
        self.flow_page = None
        super(FlowView, self).__init__(session, view_id, hidden=hidden, area=area)

    def build_top(self, top_parent, top_layout, header_parent, header_layout):
        self.add_header_tool('*', '*', 'Duplicate View', self.create_view)

        self.view_menu.setTitle('Options')
        a = self.view_menu.addAction('Show Hidden Relations')
        a.setCheckable(True)
        a.setChecked(False)
        a.toggled.connect(self.set_show_hidden_relations)

        a = self.view_menu.addAction('Show References')
        a.setCheckable(True)
        a.setChecked(False)
        a.toggled.connect(self.set_show_references_relations)

        self.view_menu.addAction('Create New View')
        self.view_menu.addSeparator()
        self.view_menu.addAction(
            'Activate DEV Tools',
            self._activate_dev_tools
        )

        self.script_line = ScriptLine(top_parent, self)
        self.script_line.hide()
        top_layout.addWidget(self.script_line, 100)

    def build_page(self, main_parent):
        self.flow_page = FlowPage(
            main_parent, self, self._start_oid, self._root_oid
        )

        lo = QtWidgets.QVBoxLayout()
        lo.setContentsMargins(0, 0, 0, 0)
        lo.addWidget(self.flow_page)
        self.flow_page.show()

        main_parent.setLayout(lo)
        self.flow_page.refresh()

    def _build(self, top_parent, top_layout, main_parent, header_parent, header_layout):
        self.build_top(top_parent, top_layout, header_parent, header_layout)
        self.build_page(main_parent)

    def get_view_state(self):
        self.setObjectName(self.view_id())
        self.dock_widget().setObjectName(self.view_id())
        
        return dict(
            oid=self.flow_page.current_oid(),
            root_oid=self._root_oid,
            form_config=self.flow_page.form.config()
        )

    def set_view_state(self, state):
        oid = state.get('oid', self.flow_page.current_oid())
        root_oid = state.get('root_oid', self._root_oid)
        form_config = state.get('form_config')

        if form_config is not None:
            self.flow_page.form.configure(**form_config)

        self._start_oid = oid
        self._root_oid = root_oid
        self.flow_page._navigator._root_oid = root_oid # ugly !!! add a public api for that !
        self.flow_page.goto(oid)

    def set_show_references_relations(self, b):
        self.flow_page.form.configure(show_references_relation=b)

    def set_show_hidden_relations(self, b):
        self.flow_page.form.configure(show_hidden_relations=b)

    def set_show_protected_relations(self, b):
        self.flow_page.form.configure(show_protected_relations=b)

    def set_group_relations(self, b):
        self.flow_page.form.configure(group_relations=b)

    def toggle_script_line(self):
        self.script_line.setVisible(not self.script_line.isVisible())

    def reload_projects(self):
        self.session.cmds.Flow.reload_project(self.flow_page.current_oid())
        self.flow_page.refresh()

    def _activate_dev_tools(self):
        if self.dev_menu is not None:
            return
        self.dev_menu = self.menubar.addMenu('[DEV]')

        self.dev_menu.addAction('Toggle Script Line', self.toggle_script_line)

        a = self.view_menu.addAction('Group Relations')
        a.setCheckable(True)
        a.setChecked(True)
        a.toggled.connect(self.set_group_relations)

        a = self.dev_menu.addAction('Show Protected Relations')
        a.setCheckable(True)
        a.setChecked(False)
        a.toggled.connect(self.set_show_protected_relations)

        self.dev_menu.addSeparator()

        self.dev_menu.addAction('Reload Projects Definition', self.reload_projects)

        self.toggle_script_line()

    def create_view(self, oid=None):
        if oid is None:
            oid = self.flow_page.current_oid()
        self.duplicate_view(oid=oid)

    def receive_event(self, event, data):
        # print('   --- FlowView got event', event, data)
        if event == 'flow_touched':
            self.flow_page.receive_event(event, data)

    def on_action_result_close(self):
        '''
        This is called when an action show in self.flow_page is run and its result
        requests closing the containing view.
        This has meaning only for a DialogView so nothing is done here.
        '''
        pass

    def goto_request(self, oid):
        self.flow_page.goto(oid)

class FlowContextView(FlowView):

    @classmethod
    def get_context_oid(cls):
        return os.environ.get('KABARET_CONTEXT_OID')

    @classmethod
    def view_type_name(cls):
        return 'Context'

    #def __init__(self, session, view_id=None, hidden=False, area=None, oid=None, root_oid=None):

    def __init__(
        self, session, view_id=None, hidden=False, area=None,
    ):
        self._context_oid = self.get_context_oid()
        super(FlowContextView, self).__init__(
            session, view_id=view_id, hidden=hidden, area=area,
            oid=self._context_oid, root_oid=self._context_oid
        )

        self.set_view_title("Kabaret Context")

    def create_view(self, oid=None):
        if oid is None:
            oid = self.get_context_oid()
        self.duplicate_view(oid=oid)


from __future__ import print_function

import six
import json

from kabaret import flow


class RedisValueStore(flow.AbstractValueStore):

    def __init__(self, redis_db, cluster_name, project_name):
        super(RedisValueStore, self).__init__()
        self._db = redis_db
        self._cluster_name = cluster_name
        self._project_name = project_name
        self._namespace = ':'.join((
            self._cluster_name, 'Flow', self._project_name
        ))

    def _key(self, key):
        return '%s:%s' % (self._namespace, key)

    def _d(self, value):
        return json.dumps(value)

    def _l(self, string):
        try:
            return json.loads(string)
        except ValueError as err:
            raise KeyError(str(err))

    def get(self, key):
        return self._l(self._db[self._key(key)])

    def set(self, key, value):
        self._db[self._key(key)] = self._d(value)

    def delete(self, key):
        try:
            del self._db[self._key(key)]
        except KeyError:
            return

    def incr(self, key, by):
        self._db.incr(self._key(key), by)

    def decr(self, key, by):
        self.incr(key, -by)

    #--- Ordererd Sting Set

    def oss_get(self, key):
        return self.oss_get_range(key, 0, -1)

    def oss_get_range(self, key, first, last):
        return self._db.zrange(self._key(key), first, last)

    def oss_has(self, key, member):
        return self._db.zscore(self._key(key), member) is not None

    def oss_add(self, key, member, score):
        self.oss_set_score(key, member, score)

    def oss_remove(self, key, member):
        return self._db.zrem(self._key(key), member)

    def oss_len(self, key):
        return self._db.zcard(self._key(key))

    def oss_get_score(self, key, member):
        return self._db.zscore(self._key(key))

    def oss_set_score(self, key, member, score):
        self._db.zadd(self._key(key), **{member: score})

    #--- HASH

    def hash_get_key(self, key, hash_key):
        # hget
        return self._l(self._db.hget(self._key(key), hash_key))

    def hash_has_key(self, key, hash_key):
        # hexists
        return self._db.hexists(self._key(key), hash_key)
        # try:
        #     return hash_key in self.store[key]
        # except:
        #     return False

    def del_hash_key(self, key, hash_key):
        # hdel
        self._db.hdel(self._key(key), hash_key)

    def get_hash(self, key):
        # not the same as get_hash_as_dict bc it keep order
        keys = self.get_hash_keys(key)
        return [
            (k, self.hash_get_key(key, k))
            for k in keys
        ]

    def get_hash_as_dict(self, key):
        # hgetall
        d = self._db.hgetall(self._key(key))
        d = dict([
            (k, self._l(v))
            for k, v in six.iteritems(d)

        ])
        return d

    def get_hash_keys(self, key):
        # hkeys
        return self._db.hkeys(self._key(key))
        # try:
        #     return self.store[key].keys()
        # except:
        #     return []

    def get_hash_len(self, key):
        # hlen
        return self._db.hlen(self._key(key))
        # try:
        #     return len(self.store[key])
        # except KeyError:
        #     return 0

    def update_hash(self, key, mapping):
        # hmset
        mapping = [
            (k, self._d(v))
            for k, v in six.iteritems(mapping)
        ]
        self._db.hmset(self._key(key), mapping)

    def set_hash(self, key, mapping):
        k = self._key(key)
        self._db.delete(k)
        if mapping:
            # mapping can be a 2d list or a dict:
            if isinstance(mapping, dict):
                mapping = mapping.items()
            mapping = [
                (mk, self._d(v))
                for mk, v in mapping]
            # looks like redis dont want to set {} as mapping...
            self._db.hmset(k, mapping)

    def set_hash_key(self, key, hash_key, value):
        # hset
        return self._db.hset(
            self._key(key), hash_key, self._d(value)
        )





from . import styles 			# init the style manager
from .styles import neonative, dark

neonative.NeonativeStyle()	# registers the style in the StyleManager
dark.DarkStyle()     		# registers the style in the StyleManager, last one is default one.


from .standalone import KabaretStandaloneGUISession
from .embedded import KabaretEmbeddedGUISession




from kabaret.flow import (
    values, group,
    Object, Map, Action, ConnectAction,
    ChoiceValueSetAction, ChoiceValueSelectAction,
    Child, Parent, Computed, Connection,
    Param, IntParam, BoolParam, HashParam,
    Label, Separator,
)


class ParentExampleChild(Object):

    my_parent = Parent()
    relations = Parent(2)


class ParentExample(Object):

    relations = Parent()
    my_child = Child(ParentExampleChild)

    doc = Label(
        '<font color=#F88>/!\\ BEWARE:</font>'
        ' When your Objects need to access other objects, it is better to use a '
        'Connection or a method returning the object to use.<br>'
        'By doing so, you make your Object more reusable.'
        'The Parent() relation is usefull to add a button to the GUI, or when the relation '
        'is strong enough to not need an overrid in subclasses...'

    )


class MyChild(Object):

    doc = Label(
        'Some pretty dumb Object...'
    )

    def summary(self):
        return 'This is a child object. Double click its name to enter it, double click here to open it.'


class Option(Object):

    @classmethod
    def get_source_display(cls, oid):
        '''
        This is used by the GUI to display the source of a Connection.
        Default is to show the whole oid. Here we cut to the source label.
        '''
        return oid.split('/')[-1].replace('_', ' ').title()


class Option1(Option):

    ICON = 'maya'

    doc = Label('This is probably not a good option...')

class Option2(Option):

    ICON = '3dsmax'

    doc = Label('Is this even en option?')


class ConnectionExample(Object):

    option_maya = Child(Option1)
    option_3ds = Child(Option2)
    current = Connection(Option)


class RelationsGroup(Object):

    doc = Label(
        '<HR><H2>'
        'Relation is what binds objects together.<br>'
        'There are several relations types, all but the "Param" one are showcased here.<br>'
        '(See the Value page for examples of the Param relations)'
        '</H2>'
    )

    child_doc = Label(
        '<HR><H3>'
        'The <u>Child</u> relation let you add an Object under another one:'

    )
    my_child = Child(MyChild)

    parent_doc = Label(
        '<HR><H3>'
        'The <u>Parent</u> relation gives a reference to an Object\'s (grand)parent.<br>'
        'The GUI will show a "goto..." button for each Parent relation of the current '
        'page Object.'
        '</H3>'
        '(Open "Parent Example" and enter "My Child" to see the buttons)'
    )
    parent_example = Child(ParentExample)

    conn_doc = Label(
        '<HR><H3>'
        'The <u>Connection</u> relation lets the user specify the related Object.<br>'
        'In this example you can drag and drop one the the two options to the "Current" relation.'
    )
    connection_example = Child(ConnectionExample)

    label_doc = Label(
        '<HR><H3>'
        'Some relation are just GUI sugar, like this "Label" for example.<br>'
        'You can also add simple separators like this one:'
    )
    sep = Separator()

    group_doc = Label(
        '<HR><H3>'
        'You can group your relations with the group context by using "with kabaret.flow.group(&lt;name&gt;):".<br>'
        'Children are grouped together'
    )
    with group('foo'):

        bar = Child(MyChild)
        baz = Child(MyChild)

        with group('yolo'):
            wip = Label(
                '<HR><H3>'
                'Nested groups feature in progress...'
            )

    sep2 = Separator()



'''
    This module defines the ViewMixin which must be used by all "View", ui or gui...
'''

import six

import uuid

class ViewMixin(object):

    @classmethod
    def view_type_name(cls):
        return cls.__name__

    def __init__(self, session, view_id=None):
        object.__init__(self)

        self.session = session
        self._view_id = view_id or str(uuid.uuid4())

        self._view_title = self.view_type_name()
        self._on_view_title_change_callback = None

        self.session.declare_view(self)

    def delete_view(self):
        self.session.forget_view(self)

    def view_id(self):
        return self._view_id

    def set_on_view_title_change(self, f):
        self._on_view_title_change_callback = f

    def set_view_title(self, title):
        self._view_title = title
        if self._on_view_title_change_callback is not None:
            self._on_view_title_change_callback()

    def view_title(self):
        return self._view_title

    def get_view_state(self):
        '''
        Subclass must override this and return a value
        suitable for set_view_stat(state) in order to support
        session managed state.

        Default value is to return {} which tells the system
        to restore the view with its default state.

        To exclude the view from being restored, you must
        override this method and return None
        '''
        return {}

    def set_view_state(self, state):
        '''
        Subclass must override this and use the state value
        returned by get_view_stat(state) in order to support
        session managed state.
        '''
        return 

    def receive_event(self, event, data):
        raise NotImplementedError()

    def goto_requested(self, requested):
        '''
        This is called when someting (another view, most probably)
        wants to alter this view's content.
        Subclass must implement their specific behavior.
        '''
        raise NotImplementedError()



'''


'''
import os
import math

import six
from qtpy import QtWidgets, QtGui, QtCore

from kabaret.app import resources

from .. import event_filters
from ..form_tree import FormTree
from ..editors import editor_factory

from .flow_field import FlowField, ObjectSummary, ObjectActionsMenu, ReferencesField
from .map_field import MapField


class ObjectField(FlowField):

    def activated(self, col):
        if not super(ObjectField, self).activated(col):
            # It it did not open another window:
            if col == 0:
                self.goto()
            else:
                if not self.isExpanded():
                    self.ensure_children_built()
                    self.setExpanded(True)
                else:
                    self.setExpanded(False)

    def goto(self):
        self.treeWidget().goto(self.oid)

    def build(self):
        # we may have 0 buttons, so ensure a default minimum size:
        self.setSizeHint(1, QtCore.QSize(32, 32))

        self.setFlags(
            QtCore.Qt.ItemIsEnabled |
            QtCore.Qt.ItemIsSelectable |
            QtCore.Qt.ItemIsDragEnabled
        )
        tree = self.treeWidget()

        self.setText(0, self._label)
        self.setIcon(0, self._get_config_icon())

        w = QtWidgets.QWidget(tree)
        lo = QtWidgets.QHBoxLayout()
        lo.setContentsMargins(2, 2, 2, 2)
        w.setLayout(lo)

        actions_menu = ObjectActionsMenu(
            self.session, w, tree.show_action_dialog, 'Flow.inline'
        )
        w.layout().addWidget(actions_menu)
        actions_menu.load_actions(self.oid)
        tree.setFirstItemColumnSpanned(self, False)

        self._summary = ObjectSummary(self.session, w)
        self._summary.load_summary(self.oid)
        w.layout().addWidget(self._summary)

        w.layout().addStretch(100)

        self.setItemWidget(1, w)

        self.show_may_have_children()

    def update_content(self):
        self._summary.load_summary(self.oid)


class ImageSequenceViewer(QtWidgets.QLabel):

    def __init__(self, parent):
        super(ImageSequenceViewer, self).__init__(parent)
        self.label = None
        self.path = None
        self.first = None
        self.last = None
        self.fps = None
        self._freq = None
        self._actual_fps = None
        self._freq_comp = 0
        self._played_frames = 0
        self._freq = None

        self.curr = None
        self.nb_frames = None
        self.aspect_ratio = None

        self._scrubbed_start_f = None
        self._scrubbed_start_x = None

        self._px_cache = {}
        self._px_height = None

        self._is_seq = False
        self._play = False

        self._last_time = QtCore.QTime()
        self._ave_time = QtCore.QTime()

        self.no_cache_brush = QtGui.QBrush(QtGui.QColor('#440000'))
        self.curr_frame_brush = QtGui.QBrush(QtGui.QColor('#00FFFF'))

        self.setFocusPolicy(QtCore.Qt.ClickFocus)

    def clear_cache(self):
        self._px_cache.clear()

    def set_sequence(self, label, path, first, last, fps):
        label = label or os.path.basename(path).split('.', 1)[0]

        self._is_seq = None not in (first, last)

        clear_cache = False
        if self.label != label:
            self.label = label
            clear_cache = True

        if self.path != path:
            self.path = path
            clear_cache = True

        if self.first != first:
            self.first = first
            clear_cache = True

        if self.last != last:
            self.last = last
            clear_cache = True

        if self.fps != fps:
            self.fps = fps
            if self.fps:
                self._freq = 1000.0 / self.fps
            clear_cache = True

        if clear_cache:
            self.clear_cache()
            if self._is_seq:
                self.nb_frames = (self.last - self.first)
                curr = self.first + self.nb_frames / 2
            else:
                self.nb_frames = 0
                curr = None
        else:
            curr = self.curr

        self.curr = None
        self.set_current_frame(curr)

    def set_current_frame(self, f):
        if not self.path:
            return
        if self._is_seq:
            f = min(self.last, max(self.first, f))
            if f == self.curr:
                return
            self.curr = f
        self._update_image()

    def _update_image(self):
        if self.path is None:
            raise ValueError('Sequence Path not set !')

        try:
            px = self._px_cache[self.curr]
        except KeyError:
            if not self._is_seq:
                p = self.path
            else:
                p = self.path % (self.curr,)
            px = QtGui.QPixmap(p)
            h = self.height()
            # pxh = px.height()
            height = h  # min(h, pxh)
            px = px.scaledToHeight(height)
            self._px_cache[self.curr] = px

        if self.aspect_ratio is None:
            self.aspect_ratio = px.width() / (px.height() * 1.0)
        self.setPixmap(px)

    def toggle_play(self):
        self._nb_avr_frames = 2
        self._actual_fps = None
        self._played_frames = 0
        if not self._play:
            self._last_time.start()
            self._ave_time.start()
            self._play = True
            self._play_step()
        else:
            self._play = False

    def _play_step(self):
        if not self._is_seq:
            self.set_current_frame(None)
            return

        elapsed = self._last_time.restart() * 1.0
        self._played_frames += 1
        if self._played_frames == self._nb_avr_frames:
            d = self._ave_time.restart() * 1.0
            self._actual_fps = 1.0 / ((d / self._nb_avr_frames) / 1000.0)
            self._played_frames = 0
            self._nb_avr_frames = min(self.fps, self._nb_avr_frames * 2)

        if self.curr == self.last:
            f = self.first
        else:
            f = self.curr + 1
        self.set_current_frame(f)
        if self._play:
            if elapsed / self._freq > 2:
                d = 0
                self._nb_avr_frames = 2
            else:
                late = elapsed - self._freq
                self._freq_comp += math.copysign(.5, late)
                d = max(0, self._freq - self._freq_comp)
            QtCore.QTimer.singleShot(max(0, d), self._play_step)

    def resizeEvent(self, e):
        self.clear_cache()
        f = self.curr
        if not self._is_seq or f is not None:
            self.curr = None
            self.set_current_frame(f)

    def keyPressEvent(self, e):
        k = e.key()
        if k == QtCore.Qt.Key_Left:
            self._play = False
            f = self.curr
            self.set_current_frame(f - 1)

        elif k == QtCore.Qt.Key_Right:
            self._play = False
            f = self.curr
            self.set_current_frame(f + 1)

        elif k == QtCore.Qt.Key_Space:
            self.toggle_play()

    def mousePressEvent(self, e):
        if e.button() == QtCore.Qt.LeftButton:
            self._scrubbed_start_f = None
            self._scrubbed_start_x = None
            e.accept()
        else:
            super(ImageSequenceViewer, self).mousePressEvent(e)

    def mouseMoveEvent(self, e):
        if e.buttons() & QtCore.Qt.LeftButton:
            if self._scrubbed_start_f is None:
                self._scrubbed_start_f = self.curr
                self._scrubbed_start_x = e.pos().x()
            else:
                self._play = False
                new_frame = self._scrubbed_start_f + (
                    e.pos().x() -
                    self._scrubbed_start_x
                ) / 1
                self.set_current_frame(new_frame)
            e.accept()
        else:
            super(ImageSequenceViewer, self).mouseMoveEvent(e)

    def mouseReleaseEvent(self, e):
        if e.button() == QtCore.Qt.LeftButton:
            self._scrubbed_start_f = None
            self._scrubbed_start_x = None
            e.accept()
        else:
            super(ImageSequenceViewer, self).mouseReleaseEvent(e)

    def paintEvent(self, e):
        super(ImageSequenceViewer, self).paintEvent(e)

        curr = self.curr
        is_seq = self._is_seq

        p = QtGui.QPainter(self)

        p.setBrush(self.curr_frame_brush)
        p.setPen(QtCore.Qt.black)
        x = 20
        y = 30
        if is_seq:
            fps = ''
            if self._actual_fps:
                fps = ' (%i fps) ' % (self._actual_fps,)
                if abs(self._actual_fps - self.fps) > 1:
                    p.setPen(QtCore.Qt.red)
            t = '%04i/%04i/%04i - %s%s' % (
                self.first, curr, self.last, self.label, fps
            )
        else:
            t = self.label

        p.drawText(x, y, t)
        p.setPen(QtCore.Qt.white)
        p.drawText(x - 1, y - 1, t)

        if not is_seq:
            return

        nb = self.nb_frames
        cr = self.contentsRect()
        fw = cr.width() / (nb * 1.0)
        h = cr.height()

        p.setPen(QtGui.QPen(QtCore.Qt.NoPen))
        p.setBrush(self.no_cache_brush)
        fr = QtCore.QRectF(0, h - 8, fw, 4)
        for i, f in enumerate(range(self.first, self.last + 1)):
            fr.moveLeft(i * fw)
            if f == curr:
                p.setBrush(self.curr_frame_brush)
                p.drawRect(fr)
                p.setBrush(self.no_cache_brush)
                continue

            try:
                self._px_cache[f]
            except KeyError:
                p.drawRect(fr)


class ObjectPreviewFied(FlowField):

    def activated(self, col):
        self.treeWidget().goto(self.oid)

    def build(self):
        self.setFlags(
            QtCore.Qt.ItemIsEnabled |
            QtCore.Qt.ItemIsSelectable |
            QtCore.Qt.ItemIsDragEnabled
        )
        tree = self.treeWidget()

        self.setText(0, self._label)

        w = QtWidgets.QWidget(tree)

        lo = QtWidgets.QHBoxLayout()
        lo.setContentsMargins(2, 2, 2, 2)
        w.setLayout(lo)

        self._thumbnail = ImageSequenceViewer(w)
        self._resizer = event_filters.MouseResizer(
            self._thumbnail, self.update_height,
            aspect_ratio=None,  # will be set in update()
        )
        self._resizer.min_h = 50
        self._resizer.propagate_events = False
        self._thumbnail.installEventFilter(self._resizer)
        # ! This connection must not be done in the ImageSequenceViewer
        # or the slot is never called !
        self._thumbnail.destroyed.connect(self._on_thumbnail_destroyed)

        lo.addWidget(self._thumbnail, 100)

        self.actions_menu = ObjectActionsMenu(
            self.session, w, tree.show_action_dialog, 'Flow.inline'
        )
        w.layout().addWidget(self.actions_menu)

        self.summary = ObjectSummary(self.session, w, self._get_icon_for)
        w.layout().addWidget(self.summary)

        w.layout().addStretch()

        self.setItemWidget(1, w)

        self.update_content()

    def _on_thumbnail_destroyed(self):
        self._thumbnail.clear_cache()

    def update_content(self):
        info = self.session.cmds.Flow.get_thumbnail_info(self.oid)
        if info.get('is_sequence'):
            label = info.get('label')
            path = info.get('path')
            first = info.get('first')
            last = info.get('last')
            fps = info.get('fps')
            default_height = info.get('default_height', 100)

        elif info.get('is_image'):
            label = info.get('label')
            path = info.get('path')
            first, last, fps = None, None, None
            default_height = info.get('default_height', 100)

        else:
            label = info.get('label')
            folder = info.get('folder')
            name = info.get('name')
            path = resources.get(folder, name)
            if path is None:
                path = resources.get('icons.gui', 'remove-symbol')
            first, last, fps = None, None, None
            default_height = info.get('default_height', 100)

        self._thumbnail.set_sequence(label, path, first, last, fps)
        if self._resizer.aspect_ratio is None:
            self._resizer.aspect_ratio = self._thumbnail.aspect_ratio
            self._resizer.apply_resize(default_height, default_height, 0, 0)

        self.actions_menu.load_actions(self.oid)

        self.summary.load_summary(self.oid)


class ParentField(FlowField):

    def __init__(self, parent, session, oid, ui_config=None):
        self._dont_build = True  # Ouch :/ uggly
        super(ParentField, self).__init__(parent, session, oid, ui_config)

        # all this uggly trick is to have the ParentFields
        # at the top of the relation list.
        # TODO: maybe this should be handled by the relation indices.
        # (so that CLI does not need to deal with this too)
        self._dont_build = False
        try:
            parent = parent.invisibleRootItem()
        except:
            # this means parent is not the tree but an item.
            # cool.
            pass

        index = parent.indexOfChild(self)
        i = parent.takeChild(index)
        # FIXME: 0 should be "index of last ParentFiled":
        parent.insertChild(0, i)
        self.build()

    def get_icon(self):
        # TODO: use a cmd to get the actual icon ?
        return self._get_icon('parent')

    def goto(self):
        self.treeWidget().goto(self.oid)

    def build(self):
        if self._dont_build:
            return

        self.setFlags(QtCore.Qt.ItemIsEnabled)  # not selectable

        tree = self.treeWidget()

        w = QtWidgets.QWidget(tree)
        lo = QtWidgets.QHBoxLayout()
        lo.setContentsMargins(2, 2, 2, 2)
        w.setLayout(lo)

        b = QtWidgets.QPushButton('Go back to %s' % (self._label,), w)
        b.clicked.connect(self.goto)
        icon = self.get_icon()
        if icon is not None:
            b.setIcon(icon)
        w.layout().addWidget(b, 100)

        if 0:
            # this is too many action everywhere, let's light it up.
            actions_menu = ObjectActionsMenu(
                self.session, w, tree.show_action_dialog, 'Flow.inline'
            )
            w.layout().addWidget(actions_menu)
            actions_menu.load_actions(self.oid)

        if 0:
            # Until the summary gets touch reactive, let's not
            # show it too much :p
            summary = ObjectSummary(self.session, w)
            summary.load_summary(self.oid)
            w.layout().addWidget(summary)

        w.layout().addStretch(100)

        self.setItemWidget(1, w)


class ConnectionButton(QtWidgets.QPushButton):

    def __init__(self, session, oid, parent):
        super(ConnectionButton, self).__init__('???', parent)
        self.session = session
        self.oid = oid
        self._editable = True
        self.setAcceptDrops(True)

    def set_editable(self, b):
        self._editable = b

    def dragEnterEvent(self, event):
        if not self._editable:
            return 
        if self.session.cmds.Flow.can_handle_mime_formats(
            event.mimeData().formats()
        ):
            event.acceptProposedAction()
        else:
            return super(ConnectionButton, self).dragEnterEvent(event)

    def dragMoveEvent(self, event):
        if not self._editable:
            return 
        if self.session.cmds.Flow.can_handle_mime_formats(
            event.mimeData().formats()
        ):
            event.acceptProposedAction()
        else:
            return super(ConnectionButton, self).dragMoveEvent(event)

    def dropEvent(self, event):
        if not self._editable:
            return 
        mime_data = event.mimeData()
        md = {}
        for format in mime_data.formats():
            md[format] = mime_data.data(format).data()
        oids, urls = self.session.cmds.Flow.from_mime_data(md)

        if not oids and not urls:
            return False    # let the event propagate up

        self.session.cmds.Flow.connect(self.oid, oids, urls)
        return True


class ConnectionField(ObjectField):

    def update_content(self):
        expanded = self.isExpanded()

        source_icon = None
        try:
            self.source_oid, source_display, source_icon = self.session.cmds.Flow.get_connection(
                self.oid
            )
        except Exception as err:
            source_display = 'ERROR: %s' % (err,)
        else:
            if source_display is None:
                source_display = '<Not Set>'
        source_icon = source_icon or 'ref'

        self.setIcon(0, self._get_icon(source_icon))
        self.b.setText(source_display)

        if expanded:
            self.delete_all_children()
            self.ensure_children_built()
            self.setExpanded(True)

    def build(self):
        # we may have 0 buttons, so ensure a default minimum size:
        self.setSizeHint(1, QtCore.QSize(20, 20))

        self.setFlags(QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable)
        tree = self.treeWidget()

        self.setText(0, self._label)

        w = QtWidgets.QWidget(tree)
        lo = QtWidgets.QHBoxLayout()
        lo.setContentsMargins(2, 2, 2, 2)
        w.setLayout(lo)

        #b = QtWidgets.QPushButton(source_display, w)
        self.b = ConnectionButton(self.session, self.oid, w)
        self.b.set_editable(self.ui_config.get('editable', True))
        self.b.clicked.connect(self.goto_ref)
        icon = self._get_icon('ref')
        if icon is not None:
            self.b.setIcon(icon)
        w.layout().addWidget(self.b)

        actions_menu = ObjectActionsMenu(
            self.session, w, tree.show_action_dialog, 'Flow.inline'
        )
        w.layout().addWidget(actions_menu)
        actions_menu.load_actions(self.oid)

        summary = ObjectSummary(self.session, w)
        summary.load_summary(self.oid)
        w.layout().addWidget(summary)

        w.layout().addStretch(100)

        self.setItemWidget(1, w)

        self.update_content()

    def goto(self):
        self.goto_ref()

    def goto_ref(self):
        if self.source_oid is not None:
            self.treeWidget().goto(self.source_oid)

    def build_children(self):
        if self.source_oid is not None:
            tree = self.treeWidget()
            tree.build_children(self, self.source_oid)


class ActionField(FlowField):

    def show_action_dialog(self):
        self.treeWidget().show_action_dialog(self.oid)

    def build(self):
        self.setFlags(QtCore.Qt.ItemIsEnabled)  # not selectable

        tree = self.treeWidget()

        # self.setText(0, '   ')
        # self.setIcon(0, self._get_config_icon())

        w = QtWidgets.QWidget(tree)
        lo = QtWidgets.QHBoxLayout()
        lo.setContentsMargins(2, 2, 2, 2)
        w.setLayout(lo)

        b = QtWidgets.QPushButton('%s' % (self._label,), w)
        b.clicked.connect(self.show_action_dialog)
        icon = self._get_config_icon()
        if icon is not None:
            b.setIcon(icon)
        w.layout().addWidget(b, 100)

        actions_menu = ObjectActionsMenu(
            self.session, w, tree.show_action_dialog, 'Flow.inline'
        )
        w.layout().addWidget(actions_menu)
        actions_menu.load_actions(self.oid)

        summary = ObjectSummary(self.session, w)
        summary.load_summary(self.oid)
        w.layout().addWidget(summary)

        w.layout().addStretch(100)

        self.setItemWidget(1, w)

    def build_children(self):
        # no child browsing on actions.
        return


class ParamField(FlowField):

    def __init__(self, parent, session, oid, ui_config=None):
        self._editor = None
        super(ParamField, self).__init__(parent, session, oid, ui_config)

    def _get_value_and_choices(self):
        value = self._get_value()
        choices = self.session.cmds.Flow.get_value_choices(self.oid)
        return value, choices

    def _get_value(self):
        return self.session.cmds.Flow.get_value(self.oid)

    def _set_value(self, new_value):
        self.session.cmds.Flow.set_value(
            self.oid, new_value
        )
        self.update_height(self._editor)

    def set_read_only(self):
        self._editor.configure(self._get_value, None)

    def build(self):
        self.setFlags(QtCore.Qt.ItemIsEnabled)  # not selectable

        tree = self.treeWidget()

        w = QtWidgets.QWidget(tree)
        resizer = event_filters.MouseResizer(
            w, self.update_height, horizontal=False)
        w.installEventFilter(resizer)

        lo = QtWidgets.QHBoxLayout()
        w.setLayout(lo)

        editor_type = self.ui_config.get('editor_type', None)
        self._editor = editor_factory().create(
            w, editor_type, self.ui_config
        )

        getter = self._get_value
        if self._editor.needs_choices():
            getter = self._get_value_and_choices
        self._editor.configure(getter, self._set_value, self._get_icon_for)

        self._editor.set_editable(self.ui_config.get('editable', True))

        self._editor.setToolTip(self.ui_config.get('tooltip', None))

        w.layout().addWidget(self._editor, stretch=100)

        actions_menu = ObjectActionsMenu(
            self.session, w, tree.show_action_dialog, 'Flow.inline'
        )
        w.layout().addWidget(actions_menu)
        w.layout().addStretch()
        actions_menu.load_actions(self.oid)

        if self._label:
            lo.setContentsMargins(2, 4, 2, 4)
            self.setText(0, self._label)
            self.setIcon(0, self._get_config_icon())
            self.setItemWidget(1, w)
        else:
            self.setText(0, '')
            self.setIcon(0, QtGui.QIcon())
            self.setItemWidget(0, w)

        self._editor.update()

    def update_content(self):
        self._editor.update()
        if self._label:
            self.setSizeHint(1, self._editor.sizeHint() * 1.1)
            self.update_height(self._editor)

    def apply_edit(self):
        if self._editor.is_edited():
            self._editor.apply()


class ActionButtonsField(FlowField):

    def build(self):
        self.setFlags(QtCore.Qt.ItemIsEnabled)  # not selectable

        tree = self.treeWidget()

        w = QtWidgets.QWidget(tree)
        lo = QtWidgets.QHBoxLayout()
        lo.setContentsMargins(2, 2, 2, 2)
        w.setLayout(lo)

        # build new buttons
        buttons = self.session.cmds.Flow.get_action_buttons(self.oid)
        for button in buttons:
            b = QtWidgets.QPushButton(button.title(), w)
            b.clicked.connect(
                lambda checked=None, b=button: self._on_button(b)
            )
            icon = self._get_icon_for(button)
            if icon is not None:
                b.setIcon(icon)
            lo.addWidget(b)

        # lo.addStretch(100)

        self.setItemWidget(1, w)

    def _on_button(self, button):
        self.treeWidget().run_action(self.oid, button)


class GroupField(FlowField):

    _DEFAULT_EXPANDED_STATE = False

    def __init__(self, parent, label):
        super(GroupField, self).__init__(
            parent, session=None, oid=label, ui_config={})

    def _pref_id(self):
        '''Overrides FromField._pref_id to omit hierachy (used by auto_expand())'''
        return 'Group:' + self._label

    def activated(self, col):
        if not self.isExpanded():
            self.setExpanded(True)
        else:
            self.setExpanded(False)

    def build_children(self):
        # we must not build children for Groups
        return

    def build(self):

        tree = self.treeWidget()

        self.setFlags(QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable)
        tree.setFirstItemColumnSpanned(self, True)

        label = QtWidgets.QLabel(tree)
        label.setText(self._label)
        label.setStyleSheet('border: 1px solid palette(mid); padding: 5px; ')
        self.setItemWidget(0, label)

    def _dispatch_touch_event(self, oid):
        for i in range(self.childCount()):
            c = self.child(i)
            handled = c.on_touch_event(oid)
            if handled:
                return True
        return False


class FlowForm(FormTree):

    def __init__(self, page, parent):
        super(FlowForm, self).__init__(parent)
        self.page = page
        self._config = dict(
            show_hidden_relations=False,
            show_protected_relations=False,
            show_reference_relations=False,
            group_relations=True,
        )

        self.setDragEnabled(True)

    def configure(
        self,
        show_hidden_relations=None, show_protected_relations=None,
        show_references_relation=None, group_relations=None, **kwargs
    ):
        if show_hidden_relations is not None:
            self._config['show_hidden_relations'] = show_hidden_relations
        if show_protected_relations is not None:
            self._config['show_protected_relations'] = show_protected_relations
        if group_relations is not None:
            self._config['group_relations'] = group_relations
        if show_references_relation is not None:
            self._config['show_reference_relations'] = show_references_relation
        self._config.update(kwargs)
        self.page.refresh()

    def config(self):
        return self._config.copy()

    def build_children(self, parent_item, oid=None):
        oid = oid or parent_item.oid
        self._build(oid, parent_item)

    def build_roots(self, oid):
        self._build(oid, self)

    def supportedDropActions(self):
        return QtCore.Qt.CopyAction

    def mimeData(self, items):
        mime_data = super(FormTree, self).mimeData(items)

        oids = [item.oid for item in items]
        md = self.page.session.cmds.Flow.to_mime_data(oids)
        for data_type, data in six.iteritems(md):
            mime_data.setData(data_type, data)

        return mime_data

    def _build(self, oid, parent_item):
        session = self.page.session

        relations, mapped_names = session.cmds.Flow.ls(
            oid,
            show_hidden=self._config['show_hidden_relations'],
            show_protected=self._config['show_protected_relations'],
            context='Flow.details'
        )

        last_group = None
        group_items = {None: parent_item}
        create_groups = self._config.get('group_relations', False)
        maps = []
        for (
            relation_name, relation_type,
            is_action, is_map, ui_config,
        ) in relations:
            related_oid = oid + '/' + relation_name
            if create_groups:
                group = ui_config.get('group')
                if group != last_group:
                    last_group = group
                    try:
                        group_item = group_items[group]
                    except KeyError:
                        group_item = GroupField(group_items[None], label=group)
                        group_items[group] = group_item
                    parent_item = group_item

            if is_action:
                ActionField(parent_item, session, related_oid, ui_config)

            elif is_map:
                maps.append(MapField(parent_item, session, related_oid, ui_config))

            elif relation_type in ('Param', 'Computed'):
                field = ParamField(parent_item, session,
                                   related_oid, ui_config)
                if relation_type == 'Computed':
                    field.set_read_only()
                field.update_content()

            elif relation_type == 'Child':
                ObjectField(parent_item, session, related_oid, ui_config)

            elif relation_type == 'Parent':
                if oid != self.page.current_oid():
                    continue
                ParentField(parent_item, session, related_oid, ui_config)

            elif relation_type == 'Connection':
                ConnectionField(parent_item, session, related_oid, ui_config)

            else:
                raise ValueError(
                    'Unsupported relation %r under %r' % (relation_type, oid)
                )

            # update gui at each child creation
            QtWidgets.QApplication.processEvents()

        is_action = session.cmds.Flow.is_action(oid)
        if is_action:
            ActionButtonsField(parent_item, session, oid)

        if session.cmds.Flow.is_map(oid):
            # MappedItemsFied(parent_item, session, oid, {})
            mapped_oids = session.cmds.Flow.get_mapped_oids(oid)
            for mapped_oid in mapped_oids:
                ObjectPreviewFied(parent_item, session, mapped_oid, {})

        if self._config['show_reference_relations']:
            ReferencesField(parent_item, session.cmds.Flow.refs(
                oid), self.goto_connected, self.goto)

        if len(maps) == 1:
            # Find a better way to implement the default expand in the map
            maps[0].setExpanded(True)

    def open(self, oid):
        self.page.open(oid)

    def goto(self, oid):
        self.page.goto(oid)

    def goto_connected(self, oid):
        self.page.goto_connected(oid)

    def show_action_dialog(self, action_oid):
        self.page.show_action_dialog(action_oid)

    def run_action(self, action_oid, button):
        self.apply_all_edits()
        self.page.run_action(action_oid, button)

    def fitSize(self):
        height = 0
        width = 0
        for i in range(self.columnCount()):
            self.resizeColumnToContents(i)
            width = width + self.columnWidth(i)
        for i in range(self.invisibleRootItem().childCount()):
            child = self.invisibleRootItem().child(i)
            widget = self.itemWidget(child, 0) or self.itemWidget(child, 1)
            if widget:
                widget.update()
                for j in range(child.childCount()):
                    height = height + self.visualItemRect(child.child(j)).height()
                height = height + widget.frameGeometry().height()
                width = max(width, widget.sizeHint().width())
        self.setMinimumSize(width * 1.1, height * 1.1)







class Cmd(object):

    @classmethod
    def cmd_name(cls):
        return cls.__name__.lower()

    def __init__(self, cmds):
        super(Cmd, self).__init__()
        self._cmds = cmds

    def actor(self):
        return self._cmds.actor()

    def __call__(self, *args, **kwargs):
        self._decode(*args, **kwargs)
        return self._execute()

    def _decode(self, *args, **kwargs):
        raise NotImplementedError('This is the abstract Cmd. It cannot be called.')

    def _execute(self):
        raise NotImplementedError('This is the abstract Cmd. It cannot be executed.')


class GetCmd(object):

    def __init__(self, CMD_TYPE):
        self.CMD_TYPE = CMD_TYPE

    def __get__(self, cmds, objtype=None):
        return self.CMD_TYPE(cmds)


class Cmds(object):

    @classmethod
    def cmd(cls, cmd_type):
        setattr(cls, cmd_type.cmd_name(), GetCmd(cmd_type))
        return cmd_type

    def __init__(self, actor):
        super(Cmds, self).__init__()
        self._actor = actor

    def actor(self):
        return self._actor


class Actor(object):

    def __init__(self, session):
        super(Actor, self).__init__()
        self.actor_name = self.__class__.__name__
        
        self.cmds = self._create_cmds()
        self.views = []

        self._session = session
        self._session._register_actor(self)

    def _create_cmds(self):
        '''
        Instanciate and return an appropriate Cmds subclass.
        '''
        raise NotImplementedError()

    def _load_views(self):
        '''
        This is called when the session has a GUI.
        Must make all the actor view types available in self._views
        '''
        pass

    def session(self):
        return self._session

    def log(self, *words):
        self._session.log(self.actor_name, *words)

    def on_session_connected(self):
        '''
        Subclasses may implement this to react to session connection.
        (f.e. subscribe to some channels)
        '''
        pass

    def die(self):
        pass




from kabaret.flow import (
    values,
    Object, Map, Action, ConnectAction,
    ChoiceValueSetAction, ChoiceValueSelectAction,
    Child, Parent, Computed, Connection,
    Param, IntParam, BoolParam, HashParam
)

from .icon_browser import IconBrowserGroup
from .showcase import ValuesGroup, MapsGroup, EditorsGroup, RelationsGroup, ActionsGroup


class UnittestProject(Object):

    values = Child(ValuesGroup)
    maps = Child(MapsGroup)
    relations = Child(RelationsGroup)
    editors = Child(EditorsGroup)
    actions = Child(ActionsGroup)

    icons = Child(IconBrowserGroup)



from __future__ import print_function

from qtpy import QtWidgets, QtGui, QtCore


class ScriptLine(QtWidgets.QWidget):

    def __init__(self, parent, flow_view):
        super(ScriptLine, self).__init__(parent)
        self.flow_view = flow_view

        self.le = QtWidgets.QLineEdit(self)
        self.le.returnPressed.connect(self._on_return)
        self.le.setToolTip(
            'Enter the method to call on this object, or "?" to list them.')

        help_bt = QtWidgets.QPushButton('?')
        help_bt.clicked.connect(self._on_help_bt)

        self.setLayout(QtWidgets.QHBoxLayout())
        self.layout().addWidget(self.le, 100)
        self.layout().addWidget(help_bt)

    def _on_return(self):
        return self.execute_object_call(str(self.le.text()).strip())

    def _on_help_bt(self):
        callables = self.execute_object_call('?')
        m = QtWidgets.QMenu()
        for name, specs in callables:
            a = m.addAction(name)
            a.triggered.connect(lambda checked=None, n=name, s=specs: self._use(n, s))
        m.exec_(QtGui.QCursor.pos())

    def _use(self, name, specs):
        (args, varargs, kw, defaults) = specs
        args.pop(0)  # drop the 'self'
        if not args:
            cmd = name + '()'
        else:
            cmd = '%s(%s)' % (name, ', '.join(args))

        self.le.setText(cmd)

    def execute_object_call(self, call_text):
        class CallRec(object):

            def __init__(self, name):
                super(CallRec, self).__init__()

                self.name = name
                self.args = ()
                self.kwargs = {}

            def __call__(self, *args, **kwargs):
                self.args = args
                self.kwargs = kwargs

            def __str__(self):
                return '%s(*%r, **%r)' % (self.name, self.args, self.kwargs)

        class Context(dict):

            def __init__(self):
                super(Context, self).__init__()

                self.calls = []

            def __getitem__(self, name):
                call = CallRec(name)
                self.calls.append(call)
                return call

        session = self.flow_view.session
        oid = self.flow_view.flow_page.current_oid()
        text = call_text.strip()
        for cmd in [t.strip() for t in text.split(';')]:
            if not cmd:
                continue
            if cmd == '?':
                result = session.cmds.Flow.call(oid, '?', (), {})
                for name, argspecs in result:
                    (args, varargs, kw, defaults) = argspecs
                return result

            else:
                context = Context()
                eval(cmd, context, context)
                if context.calls:
                    call = context.calls[0]
                    result = session.cmds.Flow.call(
                        oid, call.name, call.args, call.kwargs)
                    return result




from qtpy import QtWidgets

from ...session import KabaretSession

from .widgets.main_window import MainWindowManager, QtCore


class KabaretEmbeddedGUISession(KabaretSession):

    def __init__(
            self,
            session_name='Kabaret',
            tick_every_ms=100
    ):
        super(KabaretEmbeddedGUISession, self).__init__(session_name)
        self.main_window_manager = None
        self.tick_every_ms = tick_every_ms

    def find_main_window(self):
        raise NotImplementedError()

    def start(self):
        main_window = self.find_main_window()
        self.main_window_manager = MainWindowManager(
            self, main_window, embed_mode=True
        )
        self.main_window_manager.install()

        timer = QtCore.QTimer(self.main_window_manager.main_window)
        timer.timeout.connect(self.tick)
        timer.start(self.tick_every_ms)

    def register_view_types(self):
        '''
        Subclasses can register view types here.
        '''
        pass

    def _on_cluster_connected(self):
        '''
        Overridden to also dispatch a 'cluster_connection' event to GUI.
        '''
        super(KabaretEmbeddedGUISession, self)._on_cluster_connected()
        self.dispatch_event(
            'cluster_connection',
            cluter_name=self._cluster_actor.get_cluster_name()
        )

    # def dispatch_event(self, event_type, **data):
    #     if self.main_window_manager is not None:
    #         self.main_window_manager.dispatch_event(event_type, data)


class MayaEmbeddedSession(KabaretEmbeddedGUISession):

    def find_main_window(self):
        '''
        My attempt to get maya main window on all maya versions.
        But I only have 2015 right now so I'm kind of going blind 
        for other version :p
        '''
        try:
            from shiboken import wrapInstance
        except ImportError:
            from shiboken2 import wrapInstance

        import maya.OpenMayaUI as omui
        from qtpy import QtWidgets

        main_window_ptr = omui.MQtUtil.mainWindow()
        main_window = wrapInstance(
            long(main_window_ptr), QtWidgets.QMainWindow)
        return main_window



'''


'''


class Editor_Interface(object):
    '''
    This editor interface which define available behaviors for all kabaret editors

    Function:
        can_edit:           return True if the editor can handle the editor type
        needs_choices:      return True if the editor can accept only value among a choices list
        set_editable:       configure the editor according to the 'editable' property
        apply_options:      configure the editor according to ui configuration (dict of property)
        update:             update the editor value according to the getter (using fetch_value)
        get_edited_value:   return the editor value formatted for the setter

    Options: (needs to be applied by the editor parent)
        editable:   implies read-only
        tooltip:    the widget tooltip
    '''

    @classmethod
    def can_edit(cls, editor_type_name):
        '''
        Must be implemented to return True if the given editor_type_name
        matches this editor.
        '''
        raise NotImplementedError()

    def __init__(self, parent=None):
        '''
        The parent keyword is only there to be able to multiple inherit this w/ a widget (sucks)
        '''
        super(Editor_Interface, self).__init__()
        self._getter = None
        self._setter = None
        self._icon_provider = None
        self._is_edited = False

    def configure(self, getter, setter=None, icon_provider=None):
        '''
        Configures the value access for the editor.
        The getter and setter callables must accept arguments like:
            getter()
            setter(new_value)

        The editor will be editable if setter is not None.
        (using set_editable())
        '''
        self._getter = getter
        self._setter = setter
        self._icon_provider = icon_provider
        if self._setter is None:
            self.set_editable(False)

    def needs_choices(self):
        '''
        Must be overriden by editor presenting a choice of possible values.
        '''
        return False

    def get_icon_for(self, txt):
        if self._icon_provider is None:
            return None
        return self._icon_provider(txt)

    def set_editable(self, b):
        '''
        Must be implemented to prevent editing if b is False.
        Visual cue show also be given to the user.
        '''
        raise NotImplementedError()

    def is_edited(self):
        return self._is_edited

    def apply_options(self, options):
        '''
        Must be implemented to configure the editor as
        described by the options dict.
        '''
        raise NotImplementedError()

    def fetch_value(self):
        '''
        Returns the value to display using the getter used
        on last configure() call.
        '''
        if self._getter is None:
            raise Exception('Editor getter not configured. (use configure())')
        try:
            return self._getter()
        except Exception as err:
            self.show_exception(err)

    def update(self):
        '''
        Must be implemnented to show the value returned by self.fetch_value()
        Your code should call self._on_updated() at the end.
        '''
        raise NotImplementedError()

    def get_edited_value(self):
        '''
        Must be implemented to return the value currently displayed.
        '''
        raise NotImplementedError()

    def apply(self):
        '''
        Applies the value returned by self.get_edited_value() using
        the setter used on last configure() call.
        '''
        if self._setter is None:
            raise Exception('Value cannot be edited. (no setter configured)')
        try:
            self._setter(self.get_edited_value())
        except Exception as err:
            self.show_exception(err)
        else:
            self._show_applied()

    def _on_edited(self):
        '''
        Must be called when the value has been edited.
        '''
        self._is_edited = True
        self._show_edited()

    def _on_updated(self):
        self._is_edited = False
        self._show_clean()

    def _show_edited(self):
        '''
        Must be implemented to show that the displayed value
        needs to be applied.
        '''
        raise NotImplementedError()

    def _show_applied(self):
        '''
        Must be implemented to show that the displayed value
        as been saved.
        In a clean scenario, applying edits will trigger an update()
        and this state should disapear.
        If you are using the Editor without this kind of round trip,
        you can call update here.
        '''
        raise NotImplementedError()

    def _show_clean(self):
        '''
        Must be implemented to show that the displayed value is 
        up to date.
        '''
        raise NotImplementedError()

    def _show_error(self, error_message):
        '''
        Must be implemented to show that the given error occured.
        '''
        raise NotImplementedError()

    def show_exception(self, exception):
        import traceback
        err_msg = traceback.format_exc()
        self._show_error(err_msg)



'''

    kabaret.flow.relations

    Defines the relations classes used to bind objects together:
        Parent:     Returns the parent of the object.
                    If the nb_levels argument is more than 1, a grand parent
                    is returned: 
                        1 -> parent
                        2 -> parent's parent
                        3 -> parent's parent's parent
                        etc...

        Child:      Declares a Child in the Object. The Object will be the parent
                    of the created child.

        Param:      Declares a Child Value in the Object.
                    The default_value and value_type arguments control the initial data
                    returned by the child Value and the type of the child Value.
                    (a subclass of kabaret.flow.values.Value())
        
        SessionParam: Declare a Child SessionValue in the Object.
                    This value will be reset to its default at each session and
                    is not shared with other sessions.

        IntParam
        FloatParam
        StringParam
        DictParam
        OrderedStringSetParam
        HashParam:  Declares as Child the corresponding Value subclass.

        Separator:
                    A pre-configured Param acting as a separator in GUI

        Computed:   Declares a Child ComputedValue in the Object. 
                    See kabaret.flow.values.ComputedValue().

        Connection: Declares a Child Ref in the Object.
                    The ref being a Value returning another Object, this relations
                    is used to define dependencies between Objects.
                    See kabaret.flow.values.Ref()

'''
import logging
from .exceptions import WIPException

from .object import _Relation
from .values import (
    Value, SessionValue,
    IntValue, BoolValue, FloatValue, StringValue, DictValue,
    OrderedStringSetValue, HashValue,
    ComputedValue, Ref
)


class Parent(_Relation):
    '''
    The Parent relation give access to the related object's parent or grand-parents.
    '''
    _RELATION_TYPE_NAME = 'Parent'

    def __init__(self, nb_levels=1):
        super(Parent, self).__init__(None)
        self.nb_levels = nb_levels

    def __get__(self, o, t=None):
        if o is None:
            return self
        parent = o
        for i in range(self.nb_levels):
            if parent is None:
                raise Exception(
                    'Could not find a parent at level %d for %r' % (i, o._mng.oid()))
            parent = parent._mng.parent
        return parent


class Child(_Relation):
    '''
    The Child relation sets an Object as the Child of the owner of the relation
    (wich in turn becomes the parent)
    '''
    _RELATION_TYPE_NAME = 'Child'

    def _create_object(self, parent):
        try:
            related = self.related_type(parent, self.name)
        except:
            logging.getLogger('kabaret.flow').error(
                'Creating related object: %s(%r, %r)' % (
                    self.related_type.__name__, parent.oid(), self.name
                )
            )
            raise
        return related


class Param(Child):
    '''
    A Child relating to a :any:`Value` or one of its subclasses.
    '''

    _RELATION_TYPE_NAME = 'Param'

    _DEFAULT_VALUE_TYPE = Value

    def __init__(self, default_value=None, value_type=None):
        '''
        Beware of the default_value: if it is passed by reference, all
        instance of the class owning the created value will share the value!
        For example:
            class MyObject(Object):
                my_dict = Param({})

            -> all instances of MyObject will have their my_dict Value sharing
            the same dict (initialized when python executed the class definition)

        The workaround is to use a callable that creates a new instance of the value:
            class MyObject(Object):
                my_dict = Param(dict)

        '''
        if value_type is None:
            value_type = self.__class__._DEFAULT_VALUE_TYPE
        elif not issubclass(value_type, self.__class__._DEFAULT_VALUE_TYPE):
            raise TypeError(
                '%r is not a subclass or %r, needed for relation %r' % (
                    value_type, self.__class__._DEFAULT_VALUE_TYPE,
                    self
                )
            )

        super(Param, self).__init__(value_type)
        if callable(default_value):
            self.get_default_value = default_value
        else:
            self.get_default_value = lambda: default_value

        # default for Param is editable:
        self._ui['editable'] = True
        try:
            default_editor = value_type.DEFAULT_EDITOR
        except Exception:
            pass
        else:
            self._ui['editor_type'] = default_editor
        self._watched = False

    def watched(self, b=True):
        '''
        Configures the related value to be watched or not (default is False).
        Watched value call their parent's child_value_changed() when changed.
        '''
        self._watched = b
        return self

    def _create_object(self, parent):
        value = self.related_type(parent, self.name)
        value.set_default_value(self.get_default_value())
        value.set_watched(self._watched)
        return value

#---OLD
    def __set__(self, o, v):
        raise WIPException('Never used this, it is confusing.')
        # we must use get() on related values so not using set() feels strange
        # (you may end up thinking it's an attribute and not a related value...)
        value = self.get_related(o)
        value.set(v)


class SessionParam(Param):
    '''
    A Param relating to a :any:`SessionValue`.
    '''

    _DEFAULT_VALUE_TYPE = SessionValue


class IntParam(Param):
    '''
    A Param relating to an :any:`IntValue`.
    '''

    _DEFAULT_VALUE_TYPE = IntValue


class BoolParam(Param):
    '''
    A Param relating to a :any:`BoolValue`.
    '''

    _DEFAULT_VALUE_TYPE = BoolValue


class FloatParam(Param):
    '''
    A Param relating to a :any:`FloatValue`.
    '''

    _DEFAULT_VALUE_TYPE = FloatValue


class StringParam(Param):
    '''
    A Param relating to a :any:`StringValue`.
    '''

    _DEFAULT_VALUE_TYPE = StringValue


class DictParam(Param):
    '''
    A Param relating to a :any:`DictValue`.
    '''

    _DEFAULT_VALUE_TYPE = DictValue


class OrderedStringSetParam(Param):
    '''
    A Param relating to an :any:`OrderedStringSetValue`.
    '''

    _DEFAULT_VALUE_TYPE = OrderedStringSetValue

    def __init__(self, value_type=None):
        super(OrderedStringSetParam, self).__init__(None, value_type)


class HashParam(Param):
    '''
    A Param relating to a :any:`HashValue`.
    '''

    _DEFAULT_VALUE_TYPE = HashValue

    def __init__(self, value_type=None):
        super(HashParam, self).__init__({}, value_type)


def Separator():
    '''
    Returns a Param relation showing an horizontal line in GUI.
    '''
    return Param(None).ui(editor='label', text='<hr>').ui(label='')


def Label(text, label=''):
    '''
    Returns a Param relation showing a (potentially html) text in GUI.
    '''
    return Param(None).ui(editor='label', text=text).ui(label=label)


class Computed(Param):
    '''
    A Param relating to a :any:`ComputedValue`

    The value computation is delegated to the parent's compute_child_value() method.
    The 'cached' and 'store_value' constuctor arguments will configure the ComputedValue.
    (See kabaret.flow.values.ComputedValue)

    You can use a subclass of ComputedValue by specifying computed_value_type in the constructor
    '''
    _RELATION_TYPE_NAME = 'Computed'

    _DEFAULT_VALUE_TYPE = ComputedValue

    def __init__(self, cached=False, store_value=False, computed_value_type=None):
        super(Computed, self).__init__(None, value_type=computed_value_type)
        self._cached = cached
        self._store_value = store_value
        self.ui(editable=False)

    def _create_object(self, parent):
        value = self.related_type(parent, self.name)
        value.set_watched(self._watched)
        value.set_cached(self._cached)
        value.set_store_value(self._store_value)
        return value


class Connection(Child):
    '''
    A Child relating to a :any:`Ref` subclass.
    '''
    _RELATION_TYPE_NAME = 'Connection'

    def __init__(self, related_type=None, ref_type=None):
        '''
        One of related_type and ref_type must be given:
            - if ref_type is not None, it will be used to reference the connected object
            - if ref_type is None, related_type.ref_type() will be used

        Specifying the ref_type instead of the related_type can be usefull in 
        obscur situation but should be avoided unless you have a good reason.
        (I would explain but I cant remind why I used this in the past :p)
        '''
        if ref_type is None:
            if related_type is not None:
                ref_type = related_type.ref_type()
            # else:
            # This mean the type will be configured later (I hope!)
            # It it used for example to have a ref to the same class.
        super(Connection, self).__init__(ref_type)

        # default for Connection is editable:
        self._ui['editable'] = True
        self._watched = False

    def _create_object(self, parent):
        value = self.related_type(parent, self.name)
        value.set_watched(self._watched)
        return value

    def watched(self, b=True):
        '''
        Configures the related value to be watched or not (default is False).
        Watched value call their parent's child_value_changed() when changed.
        '''
        self._watched = b
        return self

#---OLD
    def __set__(self, o, v):
        raise WIPException('Never used this, it is confusing.')
        # we must use get() on related values so not using set() feels strange
        # (you may end up thinking it's an attribute and not a related value...)
        value = self.get_related(o)
        value.set(v)

#---OLD


class PipelineConnection(Connection):
    '''
    A Connection with ui(editable=False).
    '''

    def __init__(self, related_type=None, ref_type=None):
        raise WIPException('THIS IS OBSOLETE')
        super(PipelineConnection, self).__init__(related_type, ref_type)
        # a PipelineConnection is just a not GUI editable connection
        self._ui['editable'] = False



import os
import sys
from setuptools import setup, find_packages


# (
#    (Major, Minor, [Micros]),
#    [(releaselevel, serial)],
# )
#__version_info__ = ((1, 0, 0), ())
#__version_info__ = ((1, 0, 0),(rc,))
#__version_info__ = ((1, 0, 0),(rc,1))
__version_info__ = ((2, 1, 1), ())


if 0:
    long_description = (
        'VFX/Animation Studio Framework '
        'tailored for TDs/Scripters '
        'managing pipelines and workflows '
        'used by Production Managers and CGI Artists.\n\n'
        'Read the doc here: http://kabaret.readthedocs.io'
    )
else:
    readme = os.path.normpath(os.path.join(__file__, '..', 'README.rst'))
    with open(readme, 'r') as fh:
        long_description = fh.read()

    long_description += '\n\n'
    changelog = os.path.normpath(os.path.join(__file__, '..', 'CHANGELOG.md'))
    with open(changelog, 'r') as fh:
        long_description += fh.read()


def get_version():
    global __version_info__
    return (
        '.'.join(str(x) for x in __version_info__[0])
        + ''.join(str(x) for x in __version_info__[1])
    )


setup(
    name='kabaret',
    version=get_version(),
    description='VFX/Animation Studio Framework',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='http://www.kabaretstudio.com',
    author='Damien dee Coureau',
    author_email='kabaret-dev@googlegroups.com',
    license='LGPLv3+',
    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Desktop Environment :: File Managers',
        'Topic :: Multimedia :: Graphics',
        'Topic :: Office/Business :: Groupware',
        'Topic :: Office/Business :: Scheduling',

        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Information Technology',

        'Operating System :: OS Independent',

        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',

        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
    ],
    keywords='vfx animation framewok dataflow workflow asset manager production tracker',
    install_requires=[
        'redis<3.0.0',
        'six',
        'qtpy',
    ],
    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*',
    packages=find_packages('python'),
    package_dir={'': 'python'},
    package_data={
        '': ['*.css', '*.png', '*.svg', '*.gif'],
    },
)




import random

from kabaret.flow import (
    values,
    Object, Map, DynamicMap, Action, ConnectAction,
    ChoiceValueSetAction, ChoiceValueSelectAction,
    Child, Parent, Computed, Connection,
    Param, IntParam, BoolParam, HashParam,
    Label,
)


class MyDynamicItem(Object):

    pass


class MyDynamicMap(DynamicMap):
    '''
    The DynamicMap items are defined by implementing the methods:
        - mapped_names(...)
            Returns the names of the items.

        - _get_mapped_item_type(name) [Optional]
            Return the type of the item with the given name
    '''
    @classmethod
    def mapped_type(cls):
        return MyDynamicItem

    def mapped_names(self, page_num=0, page_size=None):
        self._last_rand_names = [
            'Item_%03i' % (i)
            for i in range(random.randint(1, 10))
        ]
        return self._last_rand_names

    def has_mapped_name(self, name):
        # NB: this is needed because the demo is based on a random list, which you'll
        # probably never have in real life scenarii...
        return name in self._last_rand_names

    def summary(self):
        return 'The number of items is randomized, refresh several times to witness it.'


class LongMap(DynamicMap):

    def __init__(self, parent, name):
        self.items = {
            'Kajus': 'Fry',
            'Gail': 'Jennings',
            'Angel': 'Collins',
            'Amelia': 'Southern',
            'Nazia': 'Parrish',
            'Niamh': 'Mullen',
            'Kailan': 'Odling',
            'Aniqa': 'Haines',
            'Brittany': 'Lamb',
            'Jareth': 'Rowe',
            'Elsie_May': 'Melia',
            'Noah': 'Robins',
            'Rochelle': 'Farrell',
            'Abdallah': 'Hickman',
            'Shayne': 'Rhodes',
            'Neve': 'Hill',
            'Kyle': 'Blair',
            'Ishmael': 'Hoover',
            'Izaan': 'Hull',
            'Giles': 'Mccray',
            'Sumaiya': 'Santana',
            'Efa': 'Townsend',
            'Alba': 'Gates',
            'Asha': 'Mcintyre',
            'Brogan': 'Matthams',
            'Tyrique': 'Mcdermott',
            'Idris': 'Patterson',
            'Ravi': 'Whiteley',
            'Leyton': 'Lambert',
            'Anabella': 'Weir',
            'Franciszek': 'Alcock',
            'Sami': 'Searle',
            'Patrycja': 'Rodriguez',
            'Francesca': 'Conrad',
            'Marie': 'Randall',
            'Zakariya': 'Baldwin',
            'Mathew': 'Gay',
            'Elysia': 'Bateman',
            'Marley': 'Simons'
        }
        super(LongMap, self).__init__(parent, name)

    def mapped_names(self, page_num=0, page_size=None):
        return self.items.keys()

    def has_mapped_name(self, name):
        return name in self.items

    def columns(self):
        return ['Name', 'Surname']

    def _fill_row_cells(self, row, item):
        row['Name'] = item.name()
        row['Surname'] = self.items[item.name()]


class OuichMap(DynamicMap):

    def __init__(self, parent, name):
        self.items = {
            'George': 'Monde de merde',
            'Le_temoin_professionnel': 'Entre, fouille-merde. Je vais t\'en filer, moi, '
                                       'du biscuit sur George pour ta feuille de chou',
            'Dino': 'T\'es un ouf toi, un ouf malade',
            '_': 'Je suis limite nervous breakdown',
            'Peter': 'On dit une ouiche lorraine.',
            'Steven': 'Moi je suis sur qu\'on dit "quiche". Enfin bon...',
            'L_indien': 'On va manger des chips ! T\'entends ?!? Des chips !'
                        ' C\'est tout ce que ca te fait quand je te dis qu\'on va manger des chips ?',
            'Hugues': 'On l\'a retrouve assassine un jour. Il en est mort.',
        }
        super(OuichMap, self).__init__(parent, name)

    def mapped_names(self, page_num=0, page_size=None):
        return self.items.keys()

    def has_mapped_name(self, name):
        return name in self.items

    def columns(self):
        return ['Name', 'Sentence']

    def _fill_row_cells(self, row, item):
        row['Name'] = item.name()
        row['Sentence'] = self.items[item.name()]


class MyItem(Object):
    pass


class AddItemAction(Action):

    _map = Parent()

    def needs_dialog(self):
        return False

    def run(self, button):
        item_name = 'Item_%03i' % (len(self._map) + 1,)
        self._map.add(item_name)
        self._map.touch()


class RemoveItemAction(Action):

    _map = Parent()

    def needs_dialog(self):
        return False

    def run(self, button):
        names = self._map.mapped_names()
        if not names:
            return
        name = names[-1]
        self._map.remove(name)
        self._map.touch()


class MyMap(Map):

    add_item = Child(AddItemAction)
    remove_item = Child(RemoveItemAction)

    @classmethod
    def mapped_type(cls):
        return MyItem

    def summary(self):
        return 'use actions here to add/remove items. ->'

class MapsGroup(Object):

    doc = Label(
        '<HR><H3>'
        'A "Map" has a per-instance list of related object.'
        '</H3>'
        'When modeling your flow, you will use maps to store arbitrary lists of objects.<br>'
        'This list may be dynamic to reflect existing stuff outside your Flow,<br>'
        'or user managed to grow your Flow instances.<br>'
        'Options are:'
        '<ul>'
        '   <li>sort_by (int or str) - The column with which the map is sorted</li>'
        '   <li>show_filter (bool) - display or not the filter field</li>'
        '   <li>default_height (int) - default height of the table'
        '   <li>auto_fit (bool) - adjust the column width to the content'
        '   <li>columns_width (tuple of percent) - width of columns'
        '</ul>'
        '<H3>Example:</H3>'
        '<pre>\n'
        'dynamic_map = Child(MyDynamicMap)\n'
        'sorted_map = Child(LongMap).ui(sort_by="Name", show_filter=True)\n'
        'bounded_map = Child(OuichMap).ui(default_height=150, auto_fit=False, columns_width=(30, 70))\n'
        '</pre>\n'
    )

    dynamic_map = Child(MyDynamicMap)
    managed_map = Child(MyMap)
    sorted_map = Child(LongMap).ui(sort_by="Name", show_filter=True)
    bounded_map = Child(OuichMap).ui(default_height=150, auto_fit=False, columns_width=(30, 70))




import os
import six
import getpass

from qtpy import QtWidgets, QtCore, QtGui

from .widget_view import ToolBarView
from kabaret.app import resources


LAYOUTS = {}

class ClickLabel(QtWidgets.QLabel):

    def __init__(self, on_click, *args, **kwargs):
        super(ClickLabel, self).__init__(*args, **kwargs)
        self._on_click = on_click

    def mousePressEvent(self, event):
        self._on_click()

class SessionToolBar(ToolBarView):

    def __init__(self, *args, **kwargs):
        super(SessionToolBar, self).__init__(*args, **kwargs)

        self.user_label = QtWidgets.QLabel(self)
        self.session_id_label = QtWidgets.QLabel(self)

        stretch = QtWidgets.QWidget(self)
        stretch.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding,
            QtWidgets.QSizePolicy.Preferred,
        )

        self.layouts_tb = QtWidgets.QToolButton(self)
        self.layouts_tb.setText('Layouts')
        self.layouts_tb.setPopupMode(self.layouts_tb.InstantPopup)
        self.layouts_tb.setIcon(resources.get_icon(('icons.gui','ui-layout')))
        self.layouts_tb.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.layout_menu = QtWidgets.QMenu(self.layouts_tb)
        self.layout_menu.addAction('TEST')
        self.layouts_tb.setMenu(self.layout_menu)

        self.addWidget(self.user_label)
        self.addSeparator()
        self.addWidget(self.session_id_label)
        self.addSeparator()
        self.addWidget(stretch)
        self.addSeparator()
        self.addWidget(self.layouts_tb)

        self.update_user_label()
        self.update_layout_tb()

    def update_user_label(self):
        label = ' <b><font size=+5>{}</font></b> '.format(
            getpass.getuser(),
        )
        self.user_label.setText(label)

        label=' {}[{}] '.format(
            self.session.session_name(),
            self.session.cmds.Cluster.get_cluster_name(),
        )
        self.session_id_label.setText(label)
        
        self.session_id_label.setToolTip(
            '<b>Session ID:</b> {}<br><b>Connection:</b> {}'.format(
                self.session.session_uid(),
                ' - '.join(self.session.cmds.Cluster.get_connection_info())
            )
        )

    def get_layout_presets(self):
        # Tmp until we have a User actor to store those things
        global LAYOUTS
        return LAYOUTS

    def store_layout_preset(self, name, layout_preset):
        # Tmp until we have a User actor to store those things
        global LAYOUTS
        LAYOUTS[name] = layout_preset

    def update_layout_tb(self):
        tb = self.layouts_tb
        menu = tb.menu()
        menu.clear()

        layout_icon = resources.get_icon(('icons.gui', 'ui-layout'))
        for name, layout in sorted(six.iteritems(self.get_layout_presets())):
            action = QtWidgets.QAction(layout_icon, name, menu)
            action.triggered.connect(lambda l=layout: self._on_set_layout_action(l))
            action.setObjectName('la')
            menu.addAction(action)

        sep = QtWidgets.QAction(menu)
        sep.setSeparator(True)
        sep.setObjectName('sep')
        menu.addAction(sep)
        #
        icon = resources.get_icon(('icons.gui', 'plus-symbol-in-a-rounded-black-square'))
        action = QtWidgets.QAction(icon, 'Store Current Layout', menu)
        action.triggered.connect(self._on_store_current_layout_action)
        action.setObjectName('sa')
        menu.addAction(action)

    def _on_set_layout_action(self, layout):
        # Don't delete us inside an signal handler from us:
        QtCore.QTimer.singleShot(
            100, lambda l=layout, s=self.session:s.set_views_state(l)
        )

    def _on_store_current_layout_action(self):
        dialog = QtWidgets.QDialog(self)
        dialog.setWindowTitle('Save Layout')
        dialog.setLayout(QtWidgets.QVBoxLayout())
        cb = QtWidgets.QComboBox(dialog)
        cb.addItems(['']+self.get_layout_presets().keys())
        cb.setEditable(True)
        dialog.layout().addWidget(cb)
        b = QtWidgets.QPushButton(dialog)
        b.setText('Save Layout')
        b.clicked.connect(dialog.accept)
        dialog.layout().addWidget(b)

        cancel = dialog.exec_() != dialog.Accepted
        name = cb.currentText().strip()
        dialog.deleteLater()
        if cancel or not name:
            return

        self.store_layout_preset(
            name,
            self.session.get_views_state()
        )
        self.update_layout_tb()

    def receive_event(self, event_type, data):
        pass



import sys, os
import logging
import platform
import getpass
import traceback
import argparse

import six

from .actors.cluster import Cluster
from .actors.flow import Flow


class SessionCmds(object):

    def __init__(self):
        super(SessionCmds, self).__init__()

class KabaretSession(object):

    @staticmethod
    def parse_command_line_args(args):
        '''
        Returns host, port, cluster_name found in the given args.
        If -h is found or if parsing fails, a ValueError is raised with 
        usage description.
        '''
        parser = argparse.ArgumentParser(
            description='Kabaret Session Arguments'
        )

        parser.add_argument(
            '-S', '--session', default='kabaret', dest='session_name', help='Session Name'
        )

        parser.add_argument(
            '-H', '--host', default='localhost', help='Cluster Host address'
        )
        parser.add_argument(
            '-P', '--port', default='6379', help='Cluster Port number'
        )
        parser.add_argument(
            '-C', '--cluster', default='DEFAULT_CLUSTER', dest='cluster_name', help='Cluster Name'
        )
        parser.add_argument(
            '-D', '--db', default='1', dest='db', help='Database Index'
        )
        parser.add_argument(
            '-p', '--password', default=None, dest='password', help='Database Password'
        )
        parser.add_argument(
            '-d', '--debug', default=False, action='store_const', const=True, dest='debug', help='Debug Mode'
        )

        values, remaining_args = parser.parse_known_args(args)
        return (
            values.session_name,
            values.host, values.port, values.cluster_name,
            values.db, values.password, values.debug, remaining_args
        )

    def __init__(self, session_name=None, debug=False):
        super(KabaretSession, self).__init__()
        self._session_name = session_name or self.__class__.__name__

        self._ticked = []
        self._actors = {}
        self.cmds = SessionCmds()
        self.debug_mode = debug

        # View Management
        self._view_types = {}
        self._views = {}

        self.log_formatter = logging.Formatter("%(name)s -- %(asctime)s -- %(levelname)s: %(message)s")
        self.stream_formatter = logging.Formatter("%(name)s - %(levelname)s: %(message)s")
        self.logger = logging.getLogger('kabaret')
        if debug:
            self.logger.setLevel(logging.DEBUG)
        else:
            self.logger.setLevel(logging.INFO)

        self.default_log_handler = logging.StreamHandler(sys.stdout)
        self.default_log_handler.setFormatter(self.stream_formatter)
        self.logger.addHandler(self.default_log_handler)

        self._cluster_actor = Cluster(self)     # this one is mandatory.
        self._create_actors()

    def add_log_file(self, filename, level=logging.INFO, mode="a", format=None, encoding="utf-8"):
        """
        Add a log file which will store all logging message with the given level
        :param filename
        :param level
        :param mode: "a" for appending or "w" for cleaning file before writing
        :param encoding
        :param format of the recording message, if None the default formatter is used
        :return: the created handler
        """
        handler = logging.FileHandler(filename, mode, encoding)
        if format:
            formatter = logging.Formatter(format)
        else:
            formatter = self.log_formatter
        handler.setFormatter(formatter)
        handler.setLevel(level)
        self.logger.addHandler(handler)
        return handler

    def add_log_stream(self, level=logging.INFO, stream=None):
        handler = logging.StreamHandler(stream)
        if format:
            formatter = logging.Formatter(format)
        else:
            formatter = self.log_formatter
        handler.setFormatter(formatter)
        handler.setLevel(level)
        self.logger.addHandler(handler)
        return handler

    def is_gui(self):
        return False

    def add_ticked(self, callable):
        '''
        Register a callable with no args to be called periodically.
        '''
        try:
            self._ticked.remove(callable)
        except ValueError:
            pass
        self._ticked.append(callable)

    def tick(self):
        for ticked in self._ticked:
            try:
                ticked()
            except Exception as err:
                message = '\n'.join((
                    "------------------ TRACE BEGIN ----------------",
                    traceback.format_exc(),
                    "------------------ TRACE END ------------------"
                ))
                self.log_error(
                    'Error while ticking: %s',
                    message
                )

    def _create_actors(self):
        '''
        Instanciate the session actors.
        Subclasses can override this to install customs actors or
        replace default ones.
        '''
        Flow(self)

    def session_name(self):
        return self._session_name

    def session_uid(self):
        return '%s:%s-%r@%s' % (
            getpass.getuser(),
            self._session_name,
            os.getpid(),
            platform.node(),
        )

    def _register_actor(self, actor):
        actor_name = actor.actor_name
        self.log(
            self._session_name,
            'Registering',
            repr(actor_name), 'Actor from', actor.__module__
        )
        self._actors[actor_name] = actor
        setattr(self.cmds, actor_name, actor.cmds)

    def get_actor_names(self):
        return sorted(self._actors.keys())

    def get_actor(self, actor_name):
        return self._actors[actor_name]

    #
    #       VIEW MANAGEMENT
    #
    def register_view_types(self):
        '''
        Subclasses can register view types and create defaults view here.
        Use:
            type_name = self.register_view_type(MyViewType)
        to register a view type.
        
        And optionally:
            view = self.add_view(type_name)
        to create a default view.
        '''
        pass

    def register_view_type(self, ViewType):
        view_type_name = ViewType.view_type_name()
        self._view_types[view_type_name] = ViewType
        return view_type_name

    def declare_view(self, view):
        '''
        This must be called by view instances uppon creation
        '''
        self._views[view.view_id()] = view

    def forget_view(self, view):
        '''
        This must be called by view instances uppon desctruction
        '''
        self._views.pop(view.view_id())

    def add_view(self, view_type_name, view_id=None, *view_args, **view_kwargs):
        try:
            ViewType = self._view_types[view_type_name]
        except KeyError:
            raise ValueError('Unknown view type %r (known view types are: %r)' %
                             (view_type_name, self._view_types.keys()))

        view = ViewType(self, view_id, *view_args, **view_kwargs)
        return view

    def view_type_count(self, view_type_name):
        return len([
            v for v in self._views.values()
            if v.view_type_name() == view_type_name
        ])

    def find_view(self, view_type_name=None, view_id=None, create=False, *args, **kwargs):
        '''
        Returns the first view with the specified view_type_name and/or the specified view_id.
        The view_id may be None to match any view_id.

        If create is True and no existsing view in found, create one in the given area
        using view_id, *args and **kwargs.

        If create is False, None is returned.
        '''
        if view_type_name is not None:
            try:
                self._view_types[view_type_name]
            except KeyError:
                raise ValueError(
                    'Find View: Unknown view type %r' % (
                        view_type_name,
                    )
                )

        for this_view_id, view in self._views.items():
            if view_type_name is not None and view.view_type_name() != view_type_name:
                continue
            if view_id is None or this_view_id == view_id:
                return view

        if create:
            return self.add_view(view_type_name, view_id, *args, **kwargs)

        return None

    def _get_layout_state(self):
        '''
        Subclasses with GUI must override this to return a state valid
        for self._set_layout_state(state)
        '''
        return None

    def _set_layout_state(self, state):
        '''
        Subclasses with GUI must implement this to restore the GUI
        state.

        The state argument is the return value of a call to 
        self._get_layout_state()
        '''
        raise NotImplementedError()

    def get_views_state(self):
        views = []
        for view_id, view in six.iteritems(self._views):
            view_state = view.get_view_state()
            if view_state is not None:
                views.append((
                    view.view_type_name(), 
                    view.view_id(),
                    view_state
                ))
            
        layout = self._get_layout_state()
        state = dict(views=views, layout=layout)
        return state

    def set_views_state(self, state):
        for view in self._views.values():
            view.delete_view()
        self._views.clear()

        from qtpy import QtWidgets
        QtWidgets.QApplication.processEvents()

        views = state.get('views', [])
        view_ids = []
        for view_type_name, view_id, view_state in views:
            view_ids.append(view_id)
            view = self.add_view(view_type_name, view_id)
            view.set_view_state(view_state)

        layout = state.get('layout')
        if layout is not None:
            self._set_layout_state(layout)

    #
    #       LOGGING
    #
    def log(self, context, *words):
        self._log(logging.INFO, ' '.join([str(i) for i in words]), extra={'context': context})

    def log_info(self, message, *args, **kwargs):
        self._log(logging.INFO, message, *args, **kwargs)

    def log_debug(self, message, *args, **kwargs):
        self._log(logging.DEBUG, message, *args, **kwargs)

    def log_error(self, message, *args, **kwargs):
        self._log(logging.ERROR, message, *args, **kwargs)

    def log_warning(self, message, *args, **kwargs):
        self._log(logging.WARNING, message, *args, **kwargs)

    def log_critical(self, message, *args, **kwargs):
        self._log(logging.CRITICAL, message, *args, **kwargs)

    def _log(self, level, message, *args, **kwargs):
        extra = {'user': self.session_uid().split(':')[0]}
        if 'extra' in kwargs:
            extra.update(kwargs['extra'])
            kwargs.pop('extra')
        self.logger.log(level, message, *args, extra=extra, **kwargs)

    def close(self):
        for actor in self._actors.values():
            actor.die()

    def _on_cluster_connected(self):
        for actor in self._actors.values():
            actor.on_session_connected()

    def channels_subscribe(self, **channels_callbacks):
        '''
        Register some handlers for the given channels.
        The handlers will be called with one arg, the message.
        The message is a dict like:
            {
                'channel': channel_name',
                'type': subscription_type  # 'subscribe' or 'psubscribe'
                'data': message_data,

            }
        Beware that if a string was sent as data, message_data will be the 
        byte encoded string. You need to decode it with:
            message_data.decode('utf8')
        Returns a callable w/o arg that will unregister those handlers.
        '''
        # WARNING: should'nt we use cmds only on actors ?!?
        # -> No, this it not to be used by the client code (cli, ui, gui...),
        # Only the Actors can subscibe callbacks (server side code), so there
        # is no cmd for that.
        return self._cluster_actor.channels_subscribe(**channels_callbacks)

    def broadcast(self, *words):
        # FIXME: Clarify if this should use the Cluster cmd or not. If clients
        # (cli, gui, ...) have a need for the command, etc...
        self._cluster_actor.broadcast(*words)

    def publish(self, **channels_messages):
        # FIXME: Clarify if this should use the Cluster cmd or not. If clients
        # (cli, gui, ...) have a need for the command, etc...
        self._cluster_actor.publish(**channels_messages)

    def dispatch_event(self, event_type, **data):
        '''
        Sends an event to all views.
        '''
        #self.log('Event', event_type, data)
        for view_id, view in six.iteritems(self._views):
            view.receive_event(event_type, data)

